# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import logging
import os
import pytest
import re
import tempfile
from pytest import raises
from mock import patch, call

from ...utils import swallow_logs
from ...tests.utils import assert_in
from ..base import ResourceManager
from ...cmd import Runner
from ..shell import Shell, ShellSession
from .test_session import check_session_passing_envvars


def test_shell_class():

    with patch.object(Runner, 'run', return_value='installed package') as runner, \
            swallow_logs(new_level=logging.DEBUG) as log:

        # Test running some install commands.
        config = {
            'name': 'my-shell',
            'type': 'shell'
        }
        shell = ResourceManager.factory(config)

        command = ['apt-get', 'install', 'bc']
        shell.add_command(command)
        command = ['apt-get', 'install', 'xeyes']
        shell.add_command(command)
        shell.execute_command_buffer()
        common_kwargs = dict(cwd=None, expect_fail=True, expect_stderr=True)
        calls = [
            call(['apt-get', 'install', 'xeyes'], **common_kwargs),
            call(['apt-get', 'install', 'bc'], **common_kwargs),
        ]
        runner.assert_has_calls(calls, any_order=True)
        assert_in("Running command '['apt-get', 'install', 'bc']'", log.lines)
        assert_in("Running command '['apt-get', 'install', 'xeyes']'", log.lines)


def test_source_file(resource_test_dir):

    # Create a temporary test file
    temp_file = tempfile.NamedTemporaryFile(dir=resource_test_dir)
    with temp_file as f:
        f.write("""
echo "Enabling special environment"
echo "We could even spit out an stderr output">&2
export EXPORTED_VAR="
multiline
"
export PATH=/custom:$PATH
NON_EXPORTED_VAR=2         # but may be those should be handled??
    """.encode('utf8'))
        f.flush()
        script = temp_file.name
        session = ShellSession()
        assert session.get_envvars() == {}
        new_env_diff = session.source_script(script, diff=True)
        assert 'EXPORTED_VAR' in new_env_diff
        assert new_env_diff['PATH'].startswith('/custom:')
        assert new_env_diff['EXPORTED_VAR'] == "\nmultiline\n"


def test_source_file_crash(resource_test_dir):
    temp_file = tempfile.NamedTemporaryFile(dir=resource_test_dir)
    with temp_file as f:
        f.write('exit 1'.encode('utf8'))
        f.flush()
        script = temp_file.name
        session = ShellSession()
        with raises(Exception):  # TODO: unify?
            session.source_script(script)


def test_isdir():
    session = ShellSession()
    assert not session.isdir(__file__)
    assert session.isdir("/bin")


def test_exists():
    session = ShellSession()
    (_, name) = tempfile.mkstemp()
    assert session.exists(name)
    os.remove(name)
    assert not session.exists(name)
    assert session.exists("/bin")


def test_source_file_param(resource_test_dir):
    temp_file = tempfile.NamedTemporaryFile(dir=resource_test_dir)
    with temp_file as f:
        f.write("""
if ! [ "$1" = "test" ]; then
   exit 1
fi
export EXPORTED_VAR=${1}1
NON_EXPORTED_VAR=2         # but may be those should be handled??
    """.encode('utf8'))
        f.flush()
        script = temp_file.name
        session = ShellSession()
        assert session.get_envvars() == {}
        new_env_diff = session.source_script([script, "test"])
        assert 'EXPORTED_VAR' in new_env_diff
        assert new_env_diff['EXPORTED_VAR'] == 'test1'

        new_env_diff = session.source_script([script, "test"],
                                         shell="/bin/bash")
        assert new_env_diff == {}


def test_session_passing_envvars():
    check_session_passing_envvars(ShellSession())


def test_shell_resource():

    config = {
        'name': 'test-ssh-resource',
        'type': 'shell'
    }
    resource = ResourceManager.factory(config)

    status = resource.create()
    assert re.match('\w{8}-\w{4}-\w{4}-\w{4}-\w{12}$', status['id']) is not None

    assert type(resource.connect()) == Shell
    assert resource.delete() is None
    with pytest.raises(NotImplementedError):
        resource.start()
    with pytest.raises(NotImplementedError):
        resource.stop()
    assert type(resource.connect()) == Shell

    with raises(NotImplementedError):
        resource.get_session(pty=True)
    with raises(NotImplementedError):
        resource.get_session(pty=False, shared=True)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Parsers of provenance information"""

from importlib import import_module
import abc
from six import add_metaclass
from six import string_types

from ..utils import file_basename
from ..dochelpers import exc_str
from ..support.exceptions import SpecLoadingError

_known_formats = ['reprozip', 'niceman', 'trig']
_known_extensions = {
    'yml': ['niceman', 'reprozip'],
    'trig': ['trig']
}

import logging
lgr = logging.getLogger('niceman.formats')

from ..distributions.base import EnvironmentSpec


# XXX Is just a file format Adapter which should provide us with functionality
# to load/store EnvironmentSpec
@add_metaclass(abc.ABCMeta)
class Provenance(object):
    """Base class to handle the collection and management of provenance files.
    
    Main purpose is to provide basic interface to provide adapters
    to generate our EnvironmentSpec object, and possibly later save it.
    
    Also should provide helpers such as `get_files` so we could do retracing.
    """

    def __init__(self, source):
        """
        Class constructor

        Parameters
        ----------
        source : string
            File path or URL to load from
        """

        self._src = self._load(source) if isinstance(source, string_types) else source

    @abc.abstractmethod
    def _load(self, source):
        return

    def get_environment(self):
        """Return Environment object 
        
        Returns
        -------
        EnvironmentSpec 
        """
        return EnvironmentSpec(
            base=self.get_base(),
            distributions=self.get_distributions(),
            files=self.get_files(limit='loose'),
        )

    # XXX should we rename into more obvious from_file/from_files?
    @staticmethod
    def factory(source, format='niceman'):
        """
        Factory method for creating the appropriate Provenance sub-class based
        on format type.

        Parameters
        ----------
        source : string
            File name or http endpoint containing provenance information.
        format : string
            ID of provenance format. Valid values are: "niceman", "reprozip"

        Returns
        -------
        Provenance sub-class instance
        """
        class_name = format.capitalize() + 'Provenance'
        module = import_module('niceman.formats.' + format)
        return getattr(module, class_name)(source)

    @staticmethod
    def chain_factory(sources):
        """Factory to load a chain of specifications

        Parameters
        ----------
        sources : list
            List of filenames or http endpoint containing provenance information.

        Raises
        ------
        SpecLoadingError
          if none of the known provenance backends were able to load the sources
        """
        fullspec = None
        for source in sources:
            if fullspec is not None:
                raise RuntimeError(
                    "Loading from multiple specifications is not yet supported")

            # try to guess from the source.  For now just filenames
            _, ext = file_basename(source, return_ext=True)
            candidates = _known_extensions.get(ext, _known_formats)
            for candidate in candidates:
                lgr.debug("Trying to load %s using %s", source, candidate)
                try:
                    # TODO: some will support 'chaining' where previous value
                    # of spec would be passed etc
                    fullspec = Provenance.factory(source, format=candidate)
                except Exception as exc:  # TODO: more specific etc
                    lgr.debug("Failed to load %s using %s: %s" % (
                              source, candidate, exc_str(exc)))
            if fullspec is None:
                raise SpecLoadingError(
                    "Failed to load %s using any known parser" % source)
        return fullspec

    # # @abc.abstractmethod
    # def get_operating_system(self):
    #     """
    #     Retrieve the operating system information.
    #
    #     Returns
    #     -------
    #     Dictionary containing name and version of the OS.
    #         os['name']
    #         os['version']
    #     """
    #     return

    def get_base(self):
        # Default
        return None

    # @abc.abstractmethod
    def get_distributions(self, in_value):
        """
        Retrieve the information for all the distributions recorded in the
        provenance file.

        Returns
        -------
        list
            List of Distribution sub-class objects.
        """
        raise NotImplementedError()

    # @abc.abstractmethod
    def get_files(self, limit='all'):
        """
        Retrieve list of files on the system which were mentioned.

        Parameters
        ----------
        limit : {'all', 'loose', 'packaged'}
          What files to return

        Returns
        -------
        list
            List of file names.
        """
        raise NotImplementedError()

    @classmethod
    def write(cls, output, spec):
        raise NotImplementedError("Output was not implemented for %s", cls)


# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Various supporting utilities for various distributions

"""

from ..debian import DebianReleaseSpec
from ..debian import get_spec_from_release_file
from ..debian import parse_dpkgquery_line

from niceman.tests.utils import eq_, assert_is_subset_recur


def test_get_spec_from_release_file(f=None):
    content = """\
-----BEGIN PGP SIGNED MESSAGE-----
Hash: SHA256

Origin: NeuroDebian
Label: NeuroDebian2
Suite: stretch
Codename: stretch2
Date: Thu, 15 Sep 2016 01:30:57 UTC
Architectures: i386 amd64 sparc
Components: main non-free contrib
Description: NeuroDebian repository with perspective, inofficial and backported packages -- mostly neuroscience-related
MD5Sum:
 d9650396c56a6f9521d0bbd9f719efbe 482669 main/binary-i386/Packages
 34134c9a64b847d33eeeb3cc7291f855ab9f0969e8ad7c92cd2a0c1aebc19d1e 14199 contrib/Contents-sparc.gz
-----BEGIN PGP SIGNATURE-----
Version: GnuPG v2

iEYEAREIAAYFAlfZ+dEACgkQpdMvASZJpamBowCfXOPQimiIy2wnVY5U9sLs1jSn
JZ0An0Uoocusvjco1t6RAwxt/y3lQoWV
=a3Nn
-----END PGP SIGNATURE-----
"""
    eq_(get_spec_from_release_file(content),
        DebianReleaseSpec(
            origin='NeuroDebian',
            label='NeuroDebian2',
            codename='stretch2',
            version=None,
            suite='stretch',
            date='Thu, 15 Sep 2016 01:30:57 UTC',
            components='main non-free contrib',
            architectures='i386 amd64 sparc',
        ))


def test_parse_apt_cache_show_pkgs_output():
    from ..debian import parse_apt_cache_show_pkgs_output
    txt1 = """\
Package: openssl
Status: install ok installed
Priority: optional
Section: utils
Installed-Size: 934
Maintainer: Ubuntu Developers <ubuntu-devel-discuss@lists.ubuntu.com>
Architecture: amd64
Version: 1.0.2g-1ubuntu4.5
Depends: libc6 (>= 2.15), libssl1.0.0 (>= 1.0.2g)
Suggests: ca-certificates
Conffiles:
 /etc/ssl/openssl.cnf 7df26c55291b33344dc15e3935dabaf3
Description-en: Secure Sockets Layer toolkit - cryptographic utility
 This package is part of the OpenSSL project's implementation of the SSL
 and TLS cryptographic protocols for secure communication over the
 Internet.
 .
 It contains the general-purpose command line binary /usr/bin/openssl,
 useful for cryptographic operations such as:
  * creating RSA, DH, and DSA key parameters;
  * creating X.509 certificates, CSRs, and CRLs;
  * calculating message digests;
  * encrypting and decrypting with ciphers;
  * testing SSL/TLS clients and servers;
  * handling S/MIME signed or encrypted mail.
Description-md5: 9b6de2bb6e1d9016aeb0f00bcf6617bd
Original-Maintainer: Debian OpenSSL Team <pkg-openssl-devel@lists.alioth.debian.org>

Package: openssl
Priority: standard
Section: utils
Installed-Size: 934
Maintainer: Ubuntu Developers <ubuntu-devel-discuss@lists.ubuntu.com>
Original-Maintainer: Debian OpenSSL Team <pkg-openssl-devel@lists.alioth.debian.org>
Architecture: amd64
Source: openssl-src (1.0.2g)
Version: 1.0.2g-1ubuntu4
Depends: libc6 (>= 2.15), libssl1.0.0 (>= 1.0.2g)
Suggests: ca-certificates
Filename: pool/main/o/openssl/openssl_1.0.2g-1ubuntu4_amd64.deb
Size: 492190
MD5sum: 8280148dc2991da94be5810ad4d91552
SHA1: b5326f27aae83c303ff934121dede47d9fce7c76
SHA256: e897ffc8d84b0d436baca5dbd684a85146ffa78d3f2d15093779d3f5a8189690
Description-en: Secure Sockets Layer toolkit - cryptographic utility
 This package is part of the OpenSSL project's implementation of the SSL
 and TLS cryptographic protocols for secure communication over the
 Internet.
 .
 It contains the general-purpose command line binary /usr/bin/openssl,
 useful for cryptographic operations such as:
  * creating RSA, DH, and DSA key parameters;
  * creating X.509 certificates, CSRs, and CRLs;
  * calculating message digests;
  * encrypting and decrypting with ciphers;
  * testing SSL/TLS clients and servers;
  * handling S/MIME signed or encrypted mail.
Description-md5: 9b6de2bb6e1d9016aeb0f00bcf6617bd
Bugs: https://bugs.launchpad.net/ubuntu/+filebug
Origin: Ubuntu
Supported: 5y
Task: standard, ubuntu-core, ubuntu-core, mythbuntu-frontend, mythbuntu-backend-slave, mythbuntu-backend-master, ubuntu-touch-core, ubuntu-touch, ubuntu-sdk-libs-tools, ubuntu-sdk

Package: alienblaster
Priority: extra
Section: universe/games
Installed-Size: 668
Maintainer: Ubuntu Developers <ubuntu-devel-discuss@lists.ubuntu.com>
Original-Maintainer: Debian Games Team <pkg-games-devel@lists.alioth.debian.org>
Architecture: amd64
Source: alienblaster-src
Version: 1.1.0-9
Depends: alienblaster-data, libc6 (>= 2.14), libgcc1 (>= 1:3.0), libsdl-mixer1.2, libsdl1.2debian (>= 1.2.11), libstdc++6 (>= 5.2)
Filename: pool/universe/a/alienblaster/alienblaster_1.1.0-9_amd64.deb
Size: 180278
MD5sum: e53379fd0d60e0af6304af78aa8ef2b7
SHA1: ca405056cf66a1c2ae3ae1674c22b7d24cda4986
SHA256: ff25bd843420801e9adea4f5ec1ca9656b2aeb327d8102107bf5ebbdb3046c38
Description-en: Classic 2D shoot 'em up
 Your mission is simple: Stop the invasion of the aliens and blast them!
 .
 Alien Blaster is a classic 2D shoot 'em up featuring lots of different
 weapons, special items, aliens to blast and a big bad boss.
 .
 It supports both a single player mode and a cooperative two player mode
 for two persons playing on one computer.
Description-md5: da1f8f1a6453d62874036331e075d65f
Homepage: http://www.schwardtnet.de/alienblaster/
Bugs: https://bugs.launchpad.net/ubuntu/+filebug
Origin: Ubuntu
"""
    out1 = [{'architecture': 'amd64',
             'package': 'openssl',
             'status': 'install ok installed',
             'version': '1.0.2g-1ubuntu4.5'},
            {'architecture': 'amd64',
             'source_name': 'openssl-src',
             'source_version': '1.0.2g',
             'package': 'openssl',
             'version': '1.0.2g-1ubuntu4'},
            {'architecture': 'amd64',
             'source_name': 'alienblaster-src',
             'package': 'alienblaster',
             'md5': 'e53379fd0d60e0af6304af78aa8ef2b7',
             'version': '1.1.0-9'},
            ]
    out = parse_apt_cache_show_pkgs_output(txt1)
    assert_is_subset_recur(out1, out, [dict, list])


def test_parse_apt_cache_policy_pkgs_output():
    from ..debian import parse_apt_cache_policy_pkgs_output
    txt1 = """\
afni:
  Installed: 16.2.07~dfsg.1-2~nd90+1
  Candidate: 16.2.07~dfsg.1-2~nd90+1
  Version table:
 *** 16.2.07~dfsg.1-2~nd90+1 500
        500 http://neuro.debian.net/debian stretch/contrib amd64 Packages
        100 /var/lib/dpkg/status
openssl:
  Installed: 1.0.2g-1ubuntu4.5
  Candidate: 1.0.2g-1ubuntu4.8
  Version table:
     1.0.2g-1ubuntu4.8 500
        500 http://us.archive.ubuntu.com/ubuntu xenial-updates/main amd64 Packages
     1.0.2g-1ubuntu4.6 500
        500 http://security.ubuntu.com/ubuntu xenial-security/main amd64 Packages
 *** 1.0.2g-1ubuntu4.5 100
        100 /var/lib/dpkg/status
     1.0.2g-1ubuntu4 500
        500 http://us.archive.ubuntu.com/ubuntu xenial/main amd64 Packages
python-nibabel:
  Installed: 2.1.0-1
  Candidate: 2.1.0-1
  Version table:
 *** 2.1.0-1 900
        900 http://http.debian.net/debian stretch/main amd64 Packages
        900 http://http.debian.net/debian stretch/main i386 Packages
        600 http://http.debian.net/debian sid/main amd64 Packages
        600 http://http.debian.net/debian sid/main i386 Packages
        100 /var/lib/dpkg/status
     2.1.0-1~nd90+1 500
        500 http://neuro.debian.net/debian stretch/main amd64 Packages
        500 http://neuro.debian.net/debian stretch/main i386 Packages
python-biotools:
  Installed: (none)
  Candidate: 1.2.12-2
  Version table:
     1.2.12-2 600
        600 http://http.debian.net/debian sid/main amd64 Packages
        600 http://http.debian.net/debian sid/main i386 Packages
alienblaster:
  Installed: 1.1.0-9
  Candidate: 1.1.0-9
  Version table:
 *** 1.1.0-9 500
        500 http://us.archive.ubuntu.com/ubuntu xenial/universe amd64 Packages
        500 file:/my/repo ./ Packages
        500 file:/my/repo2 ubuntu/ Packages
        100 /var/lib/dpkg/status
skype:i386:
  Installed: (none)
  Candidate: (none)
  Version table:
     4.3.0.37-1 -1
        100 /var/lib/dpkg/status
"""
    out1 = {'openssl': {'architecture': None,
             'candidate': '1.0.2g-1ubuntu4.8',
             'installed': '1.0.2g-1ubuntu4.5',
             'versions': [{'installed': None,
                           'priority': '500',
                           'sources': [{'priority': '500',
                                        'source': 'http://us.archive.ubuntu.com/ubuntu '
                                                  'xenial-updates/main amd64 '
                                                  'Packages'}],
                           'version': '1.0.2g-1ubuntu4.8'},
                          {'installed': None,
                           'priority': '500',
                           'sources': [{'priority': '500',
                                        'source': 'http://security.ubuntu.com/ubuntu '
                                                  'xenial-security/main amd64 '
                                                  'Packages'}],
                           'version': '1.0.2g-1ubuntu4.6'},
                          {'installed': '***',
                           'priority': '100',
                           'sources': [{'priority': '100',
                                        'source': '/var/lib/dpkg/status'}],
                           'version': '1.0.2g-1ubuntu4.5'},
                          {'installed': None,
                           'priority': '500',
                           'sources': [{'priority': '500',
                                        'source': 'http://us.archive.ubuntu.com/ubuntu '
                                                  'xenial/main amd64 '
                                                  'Packages'}],
                           'version': '1.0.2g-1ubuntu4'}]}}
    out = parse_apt_cache_policy_pkgs_output(txt1)
    assert_is_subset_recur(out1, out, [dict])

def test_parse_apt_cache_policy_source_info():
    from ..debian import parse_apt_cache_policy_source_info
    txt = """\
Package files:
 100 /var/lib/dpkg/status
     release a=now
 500 http://neuro.debian.net/debian xenial/non-free i386 Packages
     release o=NeuroDebian,a=xenial,n=xenial,l=NeuroDebian,c=non-free,b=i386
     origin neuro.debian.net
 500 http://neuro.debian.net/debian xenial/non-free amd64 Packages
     release o=NeuroDebian,a=xenial,n=xenial,l=NeuroDebian,c=non-free,b=amd64
     origin neuro.debian.net
 500 http://neuro.debian.net/debian data/non-free i386 Packages
     release o=NeuroDebian,a=data,n=data,l=NeuroDebian,c=non-free,b=i386
     origin neuro.debian.net
 500 http://neuro.debian.net/debian data/non-free amd64 Packages
     release o=NeuroDebian,a=data,n=data,l=NeuroDebian,c=non-free,b=amd64
     origin neuro.debian.net
 500 file:/my/repo2 ubuntu/ Packages
     release c=
 500 file:/my/repo ./ Packages
     release c=
 500 http://dl.google.com/linux/chrome/deb stable/main amd64 Packages
     release v=1.0,o=Google, Inc.,a=stable,n=stable,l=Google,c=main,b=amd64
     origin dl.google.com
 500 http://security.ubuntu.com/ubuntu xenial-security/restricted i386 Packages
     release v=16.04,o=Ubuntu,a=xenial-security,n=xenial,l=Ubuntu,c=restricted,b=i386
     origin security.ubuntu.com
 500 http://security.ubuntu.com/ubuntu xenial-security/restricted amd64 Packages
     release v=16.04,o=Ubuntu,a=xenial-security,n=xenial,l=Ubuntu,c=restricted,b=amd64
     origin security.ubuntu.com
 500 http://debproxy:9999/debian/ jessie-backports/contrib Translation-en
 100 http://debproxy:9999/debian/ jessie-backports/non-free amd64 Packages
     release o=Debian Backports,a=jessie-backports,n=jessie-backports,l=Debian Backports,c=non-free
     origin debproxy
 500 http://us.archive.ubuntu.com/ubuntu xenial-updates/universe amd64 Packages
     release v=16.04,o=Ubuntu,a=xenial-updates,n=xenial,l=Ubuntu,c=universe,b=amd64
     origin us.archive.ubuntu.com
 500 http://us.archive.ubuntu.com/ubuntu xenial-updates/multiverse i386 Packages
     release v=16.04,o=Ubuntu,a=xenial-updates,n=xenial,l=Ubuntu,c=multiverse,b=i386
     origin us.archive.ubuntu.com
Pinned packages:
"""
    out1 = {'http://neuro.debian.net/debian xenial/non-free i386 Packages':
                {'architecture': 'i386',
                 'archive': 'xenial',
                 'archive_uri': 'http://neuro.debian.net/debian',
                 'uri_suite': 'xenial',
                 'codename': 'xenial',
                 'component': 'non-free',
                 'label': 'NeuroDebian',
                 'origin': 'NeuroDebian',
                 'site': 'neuro.debian.net'
                 },
            'http://security.ubuntu.com/ubuntu xenial-security/restricted amd64 Packages':
                {'architecture': 'amd64',
                 'archive': 'xenial-security',
                 'archive_uri': 'http://security.ubuntu.com/ubuntu',
                 'uri_suite': 'xenial-security',
                 'codename': 'xenial',
                 'component': 'restricted',
                 'label': 'Ubuntu',
                 'origin': 'Ubuntu',
                 'site': 'security.ubuntu.com'
                 },
            'http://debproxy:9999/debian/ jessie-backports/contrib Translation-en':
                {'archive_uri': 'http://debproxy:9999/debian/',
                 'uri_suite': 'jessie-backports'
                 },
            'http://debproxy:9999/debian/ jessie-backports/non-free amd64 Packages':
                {'archive': 'jessie-backports',
                 'archive_uri': 'http://debproxy:9999/debian/',
                 'codename': 'jessie-backports',
                 'component': 'non-free',
                 'label': 'Debian Backports',
                 'origin': 'Debian Backports',
                 'site': 'debproxy',
                 'uri_suite': 'jessie-backports'
                 },
            }
    out = parse_apt_cache_policy_source_info(txt)
    assert_is_subset_recur(out1, out, [dict])


def test_get_apt_release_file_names():
    from ..debian import get_apt_release_file_names
    fn = get_apt_release_file_names('http://us.archive.ubuntu.com/ubuntu',
                                    'xenial-backports')
    assert "/var/lib/apt/lists/us.archive.ubuntu.com_ubuntu_dists_xenial-backports_InRelease" in fn
    assert "/var/lib/apt/lists/us.archive.ubuntu.com_ubuntu_dists_xenial-backports_Release" in fn
    fn = get_apt_release_file_names('file:/my/repo2/ubuntu',None)
    assert "/var/lib/apt/lists/_my_repo2_ubuntu_InRelease" in fn
    assert "/var/lib/apt/lists/_my_repo2_ubuntu_Release" in fn


def test_parse_dpkgquery_line():
    for line, expected in [
            ('zlib1g:i386: /lib/i386-linux-gnu/libz.so.1.2.8',
             {'name': 'zlib1g',
              'architecture': 'i386',
              'path': '/lib/i386-linux-gnu/libz.so.1.2.8',
              'pkgs_rest': None}),
            ('fail2ban: /usr/bin/fail2ban-client',
             {'name': 'fail2ban',
              'path': '/usr/bin/fail2ban-client',
              'pkgs_rest': None}),
            ('fsl-5.0-eddy-nonfree, fsl-5.0-core: /usr/lib/fsl/5.0',
             {'name': 'fsl-5.0-eddy-nonfree',
              'path': '/usr/lib/fsl/5.0',
              'pkgs_rest': ', fsl-5.0-core'}),
            ('pkg: path,with,commas',
             {'name': 'pkg',
              'path': 'path,with,commas',
              'pkgs_rest': None}),
            ('diversion by dash from: /bin/sh', None)
    ]:
        assert parse_dpkgquery_line(line) == expected



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Environment sub-class to provide management of an AWS EC2 instance."""

import attr
import boto3
import re
from os import chmod
from os.path import join
from appdirs import AppDirs
from botocore.exceptions import ClientError

import logging
lgr = logging.getLogger('niceman.resource.aws_ec2')

from .base import Resource
from ..ui import ui
from ..utils import assure_dir, attrib
from ..dochelpers import exc_str
from ..support.exceptions import ResourceError
from .ssh import SSH

@attr.s
class AwsEc2(Resource):

    # Generic properties of any Resource
    name = attrib(default=attr.NOTHING)

    # Configurable options for each "instance"
    access_key_id = attrib(
        doc="AWS access key for remote access to your Amazon subscription.")
    secret_access_key = attrib(
        doc="AWS secret access key for remote access to your Amazon subscription")
    instance_type = attrib(default='t2.micro',
        doc="The type of Amazon EC2 instance to run. (e.g. t2.medium)")  # EC2 instance type
    security_group = attrib(default='default',
        doc="AWS security group to assign to the EC2 instance.")  # AWS security group
    region_name = attrib(default='us-east-1',
        doc="AWS availability zone to run the EC2 instance in. (e.g. us-east-1)")  # AWS region
    key_name = attrib(
        doc="AWS subscription name of SSH key-pair registered.")  # Name of SSH key registered on AWS.
    key_filename = attrib(
        doc="Path to SSH private key file matched with AWS key name parameter.") # SSH private key filename on local machine.
    image = attrib(default='ami-c8580bdf',
        doc="AWS image ID from which to create the running instance")  # Ubuntu 14.04 LTS
    user = attrib(default='ubuntu',
        doc="Login account to EC2 instance.")

    # Interesting one -- should we allow for it to be specified or should
    # it just become a property?  may be base class could
    id = attrib()  # EC2 instance ID

    # TODO: shouldn't be hardcoded???
    type = attrib(default='aws-ec2')  # Resource type

    # Current instance properties, to be set by us, not augmented by user
    status = attrib()

    # Resource and AWS instance objects
    _ec2_resource = attrib()
    _ec2_instance = attrib()

    def connect(self):
        """
        Open a connection to the environment resource.
        """

        self._ec2_resource = boto3.resource(
            'ec2',
            aws_access_key_id=self.access_key_id,
            aws_secret_access_key=self.secret_access_key,
            region_name=self.region_name
        )
        instances = []
        if self.id:
            instances.append(self._ec2_resource.Instance(self.id))
        elif self.name:
            instances = self._ec2_resource.instances.filter(
                Filters=[{
                        'Name': 'tag:Name',
                        'Values': [self.name]
                    },
                    {
                        'Name': 'instance-state-name',
                        'Values': ['running']
                    }]
            )
            instances = list(instances)

        if len(instances) == 1:
            try:
                self._ec2_instance = instances[0]
                self.id = self._ec2_instance.instance_id
                self.status = self._ec2_instance.state['Name']
            except AttributeError:
                self.id = None
                self.status = None
        elif len(instances) > 1:
            raise ResourceError("Multiple container matches found")
        else:
            self.id = None
            self.status = None

    def create(self):
        """
        Create an EC2 instance.

        Returns
        -------
        dict : config and state parameters to capture in the inventory file
        """
        if self.id:
            raise ResourceError("Instance '{}' already exists in AWS subscription".format(
                self.id))

        if not self.key_name:
            self.create_key_pair()

        create_kwargs = dict(
            ImageId=self.image,
            InstanceType=self.instance_type,
            KeyName=self.key_name,
            MinCount=1,
            MaxCount=1,
            SecurityGroups=[self.security_group]
        )
        try:
            instances = self._ec2_resource.create_instances(**create_kwargs)
        except ClientError as exc:
            if re.search(
                "The key pair {} does not exist".format(self.key_name),
                str(exc)
            ):
                if not ui.yesno(
                    title="No key %s found in the "
                          "zone %s" % (self.key_name, self.region_name),
                    text="Would you like to generate a new key?"
                ):
                    raise
                self.create_key_pair(self.key_name)
                instances = self._ec2_resource.create_instances(**create_kwargs)
            if re.search(
                "parameter groupId is invalid", str(exc)
            ):
                raise ValueError("Invalid AWS Security Group: '{}'".format(
                    self.security_group))
            else:
                raise  # re-raise

        # Give the instance a tag name.
        self._ec2_resource.create_tags(
            Resources=[instances[0].id],
            Tags=[{'Key': 'Name', 'Value': self.name}]
        )

        # Save the EC2 Instance object.
        self._ec2_instance = self._ec2_resource.Instance(instances[0].id)
        self.id = self._ec2_instance.instance_id

        lgr.info("Waiting for EC2 instance %s to start running...", self.id)
        self._ec2_instance.wait_until_running(
            Filters=[
                {
                    'Name': 'instance-id',
                    'Values': [self.id]
                },
            ]
        )
        lgr.info("EC2 instance %s to start running!", self.id)

        lgr.info("Waiting for EC2 instance %s to complete initialization...",
                 self.id)
        waiter = self._ec2_instance.meta.client.get_waiter('instance_status_ok')
        waiter.wait(InstanceIds=[self.id])
        lgr.info("EC2 instance %s initialized!", self.id)
        self.status = self._ec2_instance.state['Name']
        return {
            'id': self.id,
            'status': self.status,
            'key_name': self.key_name,
            'key_filename': self.key_filename
        }

    def delete(self):
        """
        Terminate this EC2 instance in the AWS subscription.
        """
        self._ec2_instance.terminate()

    def start(self):
        """
        Start this EC2 instance in the AWS subscription.
        """
        self._ec2_instance.start()

    def stop(self):
        """
        Stop this EC2 instance in the AWS subscription.
        """
        self._ec2_instance.stop()

    def create_key_pair(self, key_name=None):
        """
        Walk the user through creating an SSH key pair that is saved to
        the AWS platform.
        """

        if not key_name:
            prompt = """\
You did not specify an EC2 SSH key-pair name to use when creating your EC2
environment.
Please enter a unique name to create a new key-pair or press [enter] to exit"""
            key_name = ui.question(prompt)

        # Check to see if key_name already exists. 3 tries allowed.
        for i in range(3):
            # The user wants to exit.
            if not key_name:
                raise SystemExit("Empty keyname was provided, exiting")

            key_pair = self._ec2_resource.key_pairs.filter(KeyNames=[key_name])
            try:
                try:
                    len(list(key_pair))
                except ClientError as exc:
                    # Catch the exception raised when there is no matching
                    # key name at AWS.
                    if "does not exist" in str(exc):
                        break
                    # We have no clue what it is
                    raise
            except Exception as exc:
                lgr.error(
                    "Caught some unknown exception while checking key %s: %s",
                    key_pair,
                    exc_str(exc)
                )
                # reraising
                raise

            if i == 2:
                raise SystemExit('That key name exists already, exiting.')
            else:
                key_name = ui.question('That key name exists already, try again')

        # Create private key file.
        basedir = join(
            AppDirs('niceman', 'niceman.org').user_data_dir, 'ec2_keys')
        assure_dir(basedir)
        key_filename = join(basedir, key_name + '.pem')

        # Generate the key-pair and save to the private key file.
        key_pair = self._ec2_resource.create_key_pair(KeyName=key_name)
        with open(key_filename, 'w') as key_file:
            key_file.write(key_pair.key_material)
        chmod(key_filename, 0o400)
        lgr.info('Created private key file %s', key_filename)

        # Save the new info to the resource. This is later picked up and
        # saved to the resource inventory file.
        self.key_name = key_name
        self.key_filename = key_filename

    def get_session(self, pty=False, shared=None):
        """
        Log into remote EC2 environment and get the command line
        """
        if not self._ec2_instance:
            self.connect()

        ssh = SSH(
            self.name,
            host=self._ec2_instance.public_ip_address,
            user=self.user,
            key_filename=self.key_filename
        )

        return ssh.get_session(pty=pty, shared=shared)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from ...formats import Provenance

import logging
from mock import MagicMock, patch

from niceman.utils import swallow_logs
from niceman.utils import items_to_dict
from niceman.tests.utils import assert_in


def test_distributions(demo1_spec):

    def mock_execute_command(command):
        if isinstance(command, list):
            if command == ['apt-cache', 'policy', 'libc6-dev:amd64']:
                return (
                    b'libc6-dev: \
                        Installed: (none) \
                        Candidate: 2.19-18+deb8u4 \
                        Version table: \
                            2.19-18+deb8u4 500 \
                            500 http://archive.ubuntu.com/ubuntu xenial/universe amd64 Packages',
                    0
                )
            if command == ['apt-cache', 'policy', 'afni:amd64']:
                return (
                    b'afni: \
                        Installed: 16.2.07~dfsg.1-2~nd90+1 \
                        Candidate: 16.2.07~dfsg.1-2~nd90+1 \
                        Version table: \
                            16.2.07~dfsg.1-2~nd90+1 500 \
                            500 http://archive.ubuntu.com/ubuntu xenial/universe amd64 Packages',
                    0
                )
            if command == ['apt-cache', 'policy', 'dcm2niix:amd64']:
                return (
                    b'dcm2niix: \
                        Installed: (none) \
                        Candidate: 1:1.0.20171017+git3-g9ccc4c0-1~nd16.04+1 \
                        Version table: \
                            1:1.0.20171017+git3-g9ccc4c0-1~nd16.04+1 500 \
                            500 http://archive.ubuntu.com/ubuntu xenial/universe amd64 Packages',
                    0
                )
        if isinstance(command, str):
            if command.startswith('grep'):
                return (None, 1)

    provenance = Provenance.factory(demo1_spec)
    distributions = provenance.get_distributions()
    distributions = items_to_dict(distributions)
    assert set(distributions.keys()) == {'conda', 'debian'}
    # Test DebianDistribution class.
    debian_distribution = distributions['debian']
    environment = MagicMock()
    environment.execute_command = mock_execute_command
    environment.exists.return_value = False

    with patch('requests.get') as requests, \
        swallow_logs(new_level=logging.DEBUG) as log:

        requests.return_value = type("TestObject", (object,), {})()
        requests.return_value.text = '<a href="/archive/debian/20171208T032012Z/dists/sid/">next change</a>'

        debian_distribution.initiate(environment)
        debian_distribution.install_packages(environment)

        assert_in("Adding Debian update to environment command list.", log.lines)
        assert_in("Adding line 'deb http://snapshot.debian.org/archive/debian/20170531T084046Z/ sid main \
contrib non-free' to /etc/apt/sources.list.d/niceman.sources.list", log.lines)
        assert_in("Adding line 'deb http://snapshot-neuro.debian.net:5002/archive/neurodebian/20171208T032012Z/ \
xenial main contrib non-free' to /etc/apt/sources.list.d/niceman.sources.list", log.lines)
        assert_in("Adding line 'deb http://snapshot-neuro.debian.net:5002/archive/neurodebian/20171208T032012Z/ \
xenial main contrib non-free' to /etc/apt/sources.list.d/niceman.sources.list", log.lines)
        assert_in('Installing libc6-dev=2.19-18+deb8u4, afni=16.2.07~dfsg.1-5~nd16.04+1, \
dcm2niix=1:1.0.20171017+git3-g9ccc4c0-1~nd16.04+1', log.lines)

    """
    no longer in that demo spec
    
    # Test PypiDistribution class.
    pypi_distribution = distributions[4]
    environment = MagicMock()

    pypi_distribution.initiate(environment)
    pypi_distribution.install_packages(environment)

    calls = [
        call.add_command(['pip', 'install', 'piponlypkg']),
    ]
    environment.assert_has_calls(calls, any_order=True)
    """


# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Support for Docker distribution(s)."""

import attr
import json
import logging

lgr = logging.getLogger('niceman.distributions.docker')

from .base import Package
from .base import Distribution
from .base import DistributionTracer
from .base import TypedList
from .base import _register_with_representer
from ..dochelpers import borrowdoc
from ..support.exceptions import CommandError
from ..utils import attrib


@attr.s(slots=True, frozen=True, cmp=False, hash=True)
class DockerImage(Package):
    """Docker image information"""
    id = attrib(default=attr.NOTHING)
    # Optional
    architecture = attrib()
    operating_system = attrib()
    docker_version = attrib()
    repo_digests = attrib()
    repo_tags = attrib()
    created = attrib()

_register_with_representer(DockerImage)

@attr.s
class DockerDistribution(Distribution):
    """
    Class to provide commands to Docker engine.
    """

    images = TypedList(DockerImage)

    def initiate(self, session):
        """
        Perform any initialization commands needed in the environment
        environment.

        Parameters
        ----------
        session : object
            The Session to work in
        """

        # Raise niceman.support.exceptions.CommandError exception if Docker 
        # engine is not to be found.
        session.execute_command(['docker', 'info'])

    def install_packages(self, session):
        """
        Install the Docker images associated to this distribution by the
        provenance into the environment.

        Parameters
        ----------
        session : object
            Session to work in
        """

        for image in self.images:

            # First, look in local Docker engine.
            try:
                session.execute_command(['docker', 'image', 'inspect',
                    image.id])
                continue
            except CommandError:
                pass

            # Then, look in repos.
            found = False
            for digest in image.repo_digests:
                try:
                    session.execute_command(['docker', 'pull', digest])
                    session.execute_command(['docker', 'tag', image.id,
                        image.repo_tags[0]])
                    found = True
                    break
                except CommandError:
                    pass

            # Can't find the image, so complain.
            if not found:
                raise CommandError(cmd='docker pull {}'.format(digest),
                    msg="Unable to locate Docker image {}".format(image.id))

_register_with_representer(DockerDistribution)


class DockerTracer(DistributionTracer):
    """Docker image tracer

    If a Docker engine is not found running in the session, the files
    are quietly passed on to the next tracer.
    """

    HANDLES_DIRS = False

    @borrowdoc(DistributionTracer)
    def identify_distributions(self, files):
        if not files:
            return

        # Punt if Docker daemon to found
        if self._session.execute_command('ps -e')[0].find('dockerd') == -1:
            return

        images = []
        remaining_files = set()

        for file in files:
            try:
                image = json.loads(self._session.execute_command(['docker',
                    'image', 'inspect', file])[0])[0]

                # Warn user if the image does not have any RepoDigest entries.
                if not image['RepoDigests']:
                    lgr.warning("The Docker image '%s' does not have any "
                        "repository IDs associated with it", file)

                images.append(DockerImage(
                    id=image['Id'],
                    architecture=image['Architecture'],
                    operating_system=image['Os'],
                    docker_version=image['DockerVersion'],
                    repo_digests=image['RepoDigests'],
                    repo_tags=image['RepoTags'],
                    created=image['Created']
                ))
            except CommandError as exc:
                if exc.stderr.startswith('Cannot connect to the Docker daemon'):
                    lgr.debug("Did not detect Docker engine: %s", exc)
                    return
                remaining_files.add(file)
            except Exception as exc:
                lgr.debug(exc)
                remaining_files.add(file)

        if not images:
            return

        dist = DockerDistribution(
            name="docker",
            images=images
        )

        yield dist, remaining_files

    @borrowdoc(DistributionTracer)
    def _get_packagefields_for_files(self, files):
        return

    @borrowdoc(DistributionTracer)
    def _create_package(self, **package_fields):
        return



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
""" niceman exceptions
"""


class CommandError(RuntimeError):
    """Thrown if a command call fails.
    """

    def __init__(self, cmd="", msg="", code=None, stdout="", stderr=""):
        RuntimeError.__init__(self, msg)
        self.cmd = cmd
        self.msg = msg
        self.code = code
        self.stdout = stdout
        self.stderr = stderr

    def __str__(self):
        to_str = "%s: " % self.__class__.__name__
        if self.cmd:
            to_str += "command '%s'" % (self.cmd,)
        if self.code:
            to_str += " failed with exitcode %d" % self.code
        to_str += "\n%s" % self.msg
        return to_str


class CommandNotAvailableError(CommandError):
    """Thrown if a command is not available due to certain circumstances.
    """
    pass


class InsufficientArgumentsError(ValueError):
    """To be raise instead of `ValueError` when use help output is desired"""
    pass


class MissingExternalDependency(RuntimeError):
    """External dependency is missing error"""

    def __init__(self, name, ver=None, msg=""):
        super(MissingExternalDependency, self).__init__()
        self.name = name
        self.ver = ver
        self.msg = msg

    def __str__(self):
        to_str = str(self.name)
        if self.ver:
            to_str += " of version >= %s" % self.ver
        to_str += " is missing."
        if self.msg:
            to_str += " %s" % self.msg
        return to_str


class DeprecatedError(RuntimeError):
    """To raise whenever a deprecated entirely feature is used"""
    def __init__(self, new=None, version=None, msg=''):
        """

        Parameters
        ----------
        new : str, optional
          What new construct to use
        version : str, optional
          Since which version is deprecated
        kwargs
        """
        super(DeprecatedError, self).__init__()
        self.version = version
        self.new = new
        self.msg = msg

    def __str__(self):
        s = self.msg if self.msg else ''
        if self.version:
            s += (" is deprecated" if s else "Deprecated") + " since version %s." % self.version
        if self.new:
            s += " Use %s instead." % self.new
        return s


class OutdatedExternalDependency(MissingExternalDependency):
    """External dependency is present but outdated"""

    def __init__(self, name, ver=None, ver_present=None, msg=""):
        super(OutdatedExternalDependency, self).__init__(name, ver=ver, msg=msg)
        self.ver_present = ver_present

    def __str__(self):
        to_str = super(OutdatedExternalDependency, self).__str__()
        to_str += ". You have version %s" % self.ver_present \
            if self.ver_present else \
            " Some unknown version of dependency found."
        return to_str


class SpecLoadingError(IOError):
    """To be raised when spec file fails to load"""
    pass


class MissingConfigError(RuntimeError):
    """To be raised when missing configuration a parameter"""
    pass


class MissingConfigFileError(RuntimeError):
    """To be raised when missing the configuration file"""
    pass


class MultipleReleaseFileMatch(RuntimeError):
    """Multiple release files were matched while retracing on Debian"""
    pass


#
# Generic resource errors
#


class ResourceError(RuntimeError):
    """To be raised when there is a problem with a niceman resource"""
    pass


class ResourceNotFoundError(ResourceError):
    """To be raised whenever specified resource was not found"""
    pass


class ResourceAlreadyExistsError(ResourceError):
    pass


class MultipleResourceMatches(ReferenceError):
    pass


# Session errors
class SessionRuntimeError(RuntimeError):
    pass

#
# SSH support errors, largely adopted from starcluster
#

class SSHError(Exception):
    """Base class for all SSH related errors"""
    def __init__(self, *args):
        self.args = args
        self.msg = args[0]

    def __str__(self):
        return self.msg

    def explain(self):
        return "%s: %s" % (self.__class__.__name__, self.msg)


class SSHConnectionError(SSHError):
    """Raised when ssh fails to to connect to a host (socket error)"""
    def __init__(self, host, port):
        self.msg = "failed to connect to host %s on port %s" % (host, port)


class SSHAuthException(SSHError):
    """Raised when an ssh connection fails to authenticate"""
    def __init__(self, user, host):
        self.msg = "failed to authenticate to host %s as user %s" % (host,
                                                                     user)


class SSHAccessDeniedViaAuthKeys(BaseException):
    """
    Raised when SSH access for a given user has been restricted via
    authorized_keys (common approach on UEC AMIs to allow root SSH access to be
    'toggled' via cloud-init)
    """
    def __init__(self, user):
        self.msg = "access for user '%s' denied via authorized_keys" % user



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Registry-like monster for now simply borrowed from bigmess/pymvpa

TODO: integration with cmdline etc
"""

__docformat__ = 'restructuredtext'

import os.path
from appdirs import AppDirs

from os.path import join as opj

from .support.configparserinc import SafeConfigParserWithIncludes

LOCATIONS_DOC = """
    1. '/etc/niceman/niceman.cfg'
    2. 'niceman/niceman.cfg' in all directories defined by $XDG_CONFIG_DIRS
       (by default: /etc/xdg/)
    3. 'niceman/niceman.cfg' in $XDG_CONFIG_HOME (by default: ~/.config/)
    4. '.niceman/niceman.cfg', relative to the current directory""".lstrip()


class ConfigManager(SafeConfigParserWithIncludes, object):
    """Central configuration registry for niceman.

    The purpose of this class is to collect all configurable settings
    used by various parts of niceman. It is fairly simple and does
    only little more than the standard Python ConfigParser. Like
    ConfigParser it is blind to the data that it stores, i.e. no type
    checking is performed.

    Configuration files (INI syntax) in multiple location are parsed
    when a class instance is created or whenever `Config.reload()` is
    called later on.  Files are read and parsed in the order described by
    `LOCATIONS_DOC`.

    Moreover, the constructor takes an optional argument with a list
    of additional file names to parse afterwards.

    In addition to configuration files, this class also looks for
    special environment variables to read settings from. Names of such
    variables have to start with `NICEMAN_` following by the an
    optional section name and the variable name itself ('_' as
    delimiter). If no section name is provided, the variables will be
    associated with section `general`. Some examples::

        NICEMAN_VERBOSE=1

    will become::

        [general]
        verbose = 1

    However, `NICEMAN_VERBOSE_OUTPUT=stdout` becomes::

        [verbose]
        output = stdout

    Any length of variable name as allowed, e.g. NICEMAN_SEC1_LONG_NAME=1
    becomes::

        [sec1]
        long name = 1

    Settings from custom configuration files (specified by the constructor
    argument) have the highest priority and override settings found in any of
    the config files read from default locations (which are themselves read in
    the order stated above -- overwriting earlier configuration settings
    accordingly). Finally, the content of any `NICEMAN_*` environment variables
    overrides any settings read from any file.
    """

    dirs = AppDirs("niceman", "niceman.org")

    # things we want to count on to be available
    _DEFAULTS = {'general': {'verbose': '1'}}

    def __init__(self, filenames=None, load_default=True):
        """Initialization reads settings from config files and env. variables.

        Parameters
        ----------
        filenames : list of filenames
        load_default: bool, optional
          Either to load default files (e.g. system/user-wide etc)
        """
        self._super.__init__()

        # set critical defaults
        for sec, params in ConfigManager._DEFAULTS.items():
            self.add_section(sec)
            for key, value in params.items():
                self.set(sec, key, value)

        self.__cfg_filenames = []
        self._load_default = load_default

        # now get the setting
        # TODO: look into loading/interaction with config being lazy or better more efficient...
        # could cut on startup time a bit.
        self.reload(filenames)

    @property
    def _super(self):
        return super(ConfigManager, self)

    def _get_default_file_candidates(self):
        # list of filenames to parse (custom plus some standard ones),
        # they are listed in the order they need to be parsed to
        # guarantee a sane configuration file cascade
        homedir = os.path.expanduser('~')  # seems to be useless ???
        cfg_file_candidates = [
            # shipped-with config
            # opj(os.path.dirname(__file__), 'niceman.cfg'),
            # system config
            '/etc/niceman/niceman.cfg']
        # XDG system config
        cfg_file_candidates.append(opj(self.dirs.site_config_dir,
                                       'niceman.cfg'))

        # XDG user config
        home_cfg_base_path = self.dirs.user_config_dir
        if os.path.isabs(home_cfg_base_path):
            cfg_file_candidates.append(opj(home_cfg_base_path, 'niceman.cfg'))

        # current dir config
        cfg_file_candidates.append(opj('.niceman', 'niceman.cfg'))
        return cfg_file_candidates

    def _get_file_candidates(self):
        cfg_file_candidates = self._get_default_file_candidates() if self._load_default else []
        # runtime config
        cfg_file_candidates += self.__cfg_filenames
        return cfg_file_candidates


    def reload(self, filenames=None):
        """Re-read settings from all configured locations.
        """
        # store additional config file names
        if not filenames is None:
            self.__cfg_filenames = filenames

        # read local and user-specific config
        self.read(self._get_file_candidates())

        # now look for variables in the environment
        pref = 'NICEMAN_'
        for var in [v for v in os.environ.keys() if v.startswith(pref)]:
            # strip leading 'NICEMAN_' and lower case entries
            svar = var[len(pref):].lower()

            # section is next element in name (or 'general' if simple name)
            if not svar.count('_'):
                sec = 'general'
            else:
                cut = svar.find('_')
                sec = svar[:cut]
                svar = svar[cut + 1:].replace('_', ' ')

            # check if section is already known and add it if not
            if not self.has_section(sec):
                self.add_section(sec)

            # set value
            self.set(sec, svar, os.environ[var])

    def get(self, section, option, default=None, **kwargs):
        """Wrapper around SafeConfigParser.get() with a custom default value.

        This method simply wraps the base class method, but adds a `default`
        keyword argument. The value of `default` is returned whenever the
        config parser does not have the requested option and/or section.
        """
        if not self.has_option(section, option):
            return default

        try:
            return self._super.get(section, option, **kwargs)
        except ValueError as e:
            # provide somewhat descriptive error
            raise ValueError(
                "Failed to obtain value from configuration for %s.%s. "
                "Original exception was: %s" % (section, option, e))

    def getpath(self, *args, **kwargs):
        """Wrapper around get to do additional path treatments such as expanduser

        See documentation for `get`
        """
        val = self.get(*args, **kwargs)
        return os.path.expanduser(val)

    def getboolean(self, section, option, default=None):
        """Wrapper around SafeConfigParser.getboolean() with a custom default.

        This method simply wraps the base class method, but adds a `default`
        keyword argument. The value of `default` is returned whenever the
        config parser does not have the requested option and/or section.
        """
        if not self.has_option(section, option):
            if isinstance(default, bool):
                return default
            else:
                # compatibility layer for py3 version of ConfigParser
                if hasattr(self, '_boolean_states'):
                    boolean_states = self._boolean_states
                else:
                    boolean_states = self.BOOLEAN_STATES  # there is no BOOLEAN_STATES ???
                if default is None or (default.lower() not in boolean_states):
                    raise ValueError('Not a boolean: %s' % default)
                return boolean_states[default.lower()]

        return self._super.getboolean(section, option)

    ##REF: Name was automagically refactored
    def get_as_dtype(self, section, option, dtype, default=None):
        """Convenience method to query options with a custom default and type

        This method simply wraps the base class method, but adds a
        `default` keyword argument. The value of `default` is returned
        whenever the config parser does not have the requested option
        and/or section.

        In addition, the returned value is converted into the
        specified `dtype`.
        """
        if not self.has_option(section, option):
            return default
        try:
            return self._super._get(section, dtype, option)
        except ValueError as e:
            # provide somewhat descriptive error
            raise ValueError(
                "Failed to obtain value from configuration for %s.%s. "
                "Original exception was: %s" % (section, option, e))



# emacs: -*- mode: python; py-indent-offset: 8; tab-width: 6; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Orchestrator sub-class to provide management of the localhost environment."""
import json
import os
from collections import defaultdict

import attr
import yaml
from niceman.resource.session import get_local_session

from niceman.distributions import Distribution, piputils
from niceman.dochelpers import exc_str
from niceman.support.exceptions import CommandError
from niceman.utils import attrib, PathRoot, is_subpath, make_tempfile

from .base import SpecObject
from .base import DistributionTracer
from .base import Package
from .base import TypedList


import logging
lgr = logging.getLogger('niceman.distributions.conda')


def get_conda_platform_from_python(py_platform):
    """
    Converts a python platform string to a corresponding conda platform

    Parameters
    ----------
    py_platform : str
        The python platform string

    Returns
    -------
    str
        The conda platform string

    """
    # Provides the conda platform mapping from conda/models/enum.py
    # Note that these are python prefixes (both 'linux2' and 'linux' from
    # python map to 'linux' in conda.)
    python_to_conda_platform_map = {
        'darwin': 'osx',
        'linux': 'linux',
        'openbsd': 'openbsd',
        'win': 'win',
        'zos': 'zos',
    }
    for k in python_to_conda_platform_map:
        if py_platform.startswith(k):
            return python_to_conda_platform_map[k]
    return None


def get_miniconda_url(conda_platform, python_version):
    """
    Gets the Miniconda install URL given the conda platform and python version

    Parameters
    ----------
    conda_platform : str
        The conda platform (e.g. "linux-64")

    python_version : str
        The python version (e.g. "2.7.1")

    Returns
    -------
    str
        The Miniconda insaller URL
    """
    if conda_platform.startswith("linux"):
        platform = "Linux"
    elif conda_platform.startswith("osx"):
        platform = "MacOSX"
    else:
        raise ValueError("Unsupported platform %s for conda installation" %
                         conda_platform)
    platform += "-x86_64" if ("64" in conda_platform) else "-x86"
    return "https://repo.continuum.io/miniconda/Miniconda%s-latest-%s.sh" \
                    % (python_version[0], platform)

@attr.s
class CondaPackage(Package):
    name = attrib(default=attr.NOTHING)
    installer = attrib()
    version = attrib()
    build = attrib()
    channel_name = attrib()
    size = attrib()
    md5 = attrib()
    url = attrib()
    location = attrib()
    editable = attrib(default=False)
    files = attrib(default=attr.Factory(list))

    _cmp_fields = ('name', 'build')

@attr.s
class CondaChannel(SpecObject):
    name = attrib(default=attr.NOTHING)
    url = attrib()


@attr.s
class CondaEnvironment(SpecObject):
    name = attrib(default=attr.NOTHING)
    path = attrib()
    packages = TypedList(CondaPackage)
    channels = TypedList(CondaChannel)


# ~/anaconda
#
# ~/anaconda3
@attr.s
class CondaDistribution(Distribution):
    """
    Class to provide Conda package management.
    """
    path = attrib()
    conda_version = attrib()
    python_version = attrib()
    platform = attrib()
    environments = TypedList(CondaEnvironment)

    _cmp_field = ('path',)

    def initiate(self, environment):
        """
        Perform any initialization commands needed in the environment.

        Parameters
        ----------
        environment : object
            The Environment sub-class object.
        """
        # TODO Move conda installation here (environment is actually session)
        return

    def install_packages(self, session=None):
        """
        Install the packages associated to this distribution by the provenance
        into the environment.

        Parameters
        ----------
        session : object
            Environment sub-class instance.

        Raises
        ------
        ValueError
            Unexpected conda platform or python version
        CommandError
            If unexpected error in install commands occurs
        """

        if not self.path:  # Permit empty conda config entry
            return

        if not session:
            session = get_local_session()

        # Use the session to make a temporary directory for our install files
        tmp_dir = session.mktmpdir()
        try:
            # Install Conda
            # See if Conda root path exists and if not, install Conda
            if not session.isdir(self.path):
                # TODO: Determine if we can detect miniconda vs anaconad
                miniconda_url = get_miniconda_url(self.platform,
                                                  self.python_version)
                session.execute_command("curl %s -o %s/miniconda.sh" %
                                        (miniconda_url, tmp_dir))
                # NOTE: miniconda.sh makes parent directories automatically
                session.execute_command("bash -b %s/miniconda.sh -b -p %s" %
                                        (tmp_dir, self.path))
            ## Update root version of conda
            session.execute_command(
               "%s/bin/conda install -y conda=%s python=%s" %
               (self.path, self.conda_version,
                self.get_simple_python_version(self.python_version)))

            # Loop through non-root packages, creating the conda-env config
            for env in self.environments:
                export_contents = self.create_conda_export(env)
                with make_tempfile(export_contents) as local_config:
                    remote_config = os.path.join(tmp_dir, env.name)
                    session.put(local_config, remote_config)
                    if not session.isdir(env.path):
                        try:
                            session.execute_command(
                                "%s/bin/conda-env create -p %s -f %s " %
                                (self.path, env.path, remote_config))
                        except CommandError:
                            # Some conda versions seg fault so try to update
                            session.execute_command(
                                "%s/bin/conda-env update -p %s -f %s " %
                                (self.path, env.path, remote_config))
                    else:
                        session.execute_command(
                            "%s/bin/conda-env update -p %s -f %s " %
                            (self.path, env.path, remote_config))

        finally:
            if tmp_dir:
                # Remove the tmp dir
                session.execute_command(["rm", "-R", tmp_dir])

        return

    @property
    def packages(self):
        return [ p for env in self.environments for p in env.packages ]

    @staticmethod
    def get_simple_python_version(python_version):
        # Get the simple python version from the conda info string
        # Specifically, pull "major.minor.micro" from the full string
        # "major.minor.micro.releaselevel.serial"
        return ".".join(python_version.split(".", 3)[:3])

    @staticmethod
    def format_conda_package(name, version=None, build=None, **_):
        # Note: Conda does not accept a build without a version
        return ("%s=%s=%s" % (name, version, build) if version and build
                 else ("%s=%s" % (name, version) if version
                       else "%s" % name))

    @staticmethod
    def format_pip_package(name, version=None, **_):
        return ("%s==%s" % (name, version) if version
                       else "%s" % name)

    @staticmethod
    def create_conda_export(env):
        # Collect the environment into a dictionary in the same manner as
        # https://github.com/conda/conda/blob/master/conda_env/env.py
        d = {}
        # TODO: The environment name should be discovered on retrace
        name = os.path.basename(os.path.normpath(env.path))
        d["name"] = name
        # Collect channels
        d["channels"] = [c.name for c in env.channels]
        # Collect packages (dependencies) with no installer
        d["dependencies"] = [
            CondaDistribution.format_conda_package(p.name, p.version, p.build)
            for p in env.packages
            if p.installer is None]
        #            p.get("name"), p.get("version"), p.get("build"))
        # Collect pip-installed dependencies
        pip_deps = [CondaDistribution.format_pip_package(p.name, p.version)
                    for p in env.packages
                    if p.installer is "pip"]
        if (pip_deps):
            d["dependencies"].append({"pip": pip_deps})
        # Add the prefix
        d["prefix"] = env.path
        # Now dump the export as a yaml file
        return yaml.safe_dump(d, default_flow_style=False)


class CondaTracer(DistributionTracer):
    """conda distributions tracer
    """

    def _init(self):
        self._get_conda_env_path = PathRoot(self._is_conda_env_path)

    def _get_packagefields_for_files(self, files):
        raise NotImplementedError("TODO")

    def _create_package(self, *fields):
        raise NotImplementedError("TODO")

    def _get_conda_meta_files(self, conda_path):
        try:
            out, _ = self._session.execute_command(
                'ls %s/conda-meta/*.json'
                % conda_path
            )
            return iter(out.splitlines())
        except Exception as exc:  # Empty conda environment (unusual situation)
            lgr.warning("Could not retrieve conda-meta files in path %s: %s",
                        conda_path, exc_str(exc))
            return iter(())

    def _get_conda_package_details(self, conda_path):
        packages = {}
        file_to_package_map = {}
        for meta_file in self._get_conda_meta_files(conda_path):
            try:
                out, err = self._session.execute_command(
                    'cat %s' % meta_file
                )
                details = json.loads(out)
#                print meta_file
#                print(json.dumps(details, indent=4))
                if "name" in details:
                    lgr.debug("Found conda package %s", details["name"])
                    # Packages are recorded in the conda environment as
                    # name=version=build
                    conda_package_name = \
                        ("%s=%s=%s" % (details["name"], details["version"],
                                       details["build"]))
                    packages[conda_package_name] = details
                    # Now map the package files to the package
                    for f in details["files"]:
                        full_path = os.path.normpath(
                            os.path.join(conda_path, f))
                        file_to_package_map[full_path] = conda_package_name
            except Exception as exc:
                lgr.warning("Could not retrieve conda info in path %s: %s",
                            conda_path,
                            exc_str(exc))

        return packages, file_to_package_map

    def _get_conda_pip_package_details(self, env_export, conda_path):
        dependencies = env_export.get("dependencies", [])

        # If there are pip dependencies, they'll be listed under a
        # {"pip": [...]} entry.
        pip_pkgs = []
        for dep in dependencies:
            if isinstance(dep, dict) and "pip" in dep:
                # Pip packages are recorded in conda exports as "name (loc)",
                # "name==version" or "name (loc)==version".
                pip_pkgs = [p.split("=")[0].split(" ")[0] for p in dep["pip"]]
                break

        if not pip_pkgs:
            return {}, {}

        pip = conda_path + "/bin/pip"
        packages, file_to_package_map = piputils.get_package_details(
            self._session, pip, pip_pkgs)
        for entry in packages.values():
            entry["installer"] = "pip"
        return packages, file_to_package_map

    def _get_conda_env_export(self, root_prefix, conda_path):
        export = {}
        try:
            # NOTE: We need to call conda-env directly.  Conda has problems
            # calling conda-env without a PATH being set...
            out, err = self._session.execute_command(
                '%s/bin/conda-env export -p %s'
                % (root_prefix, conda_path)
            )
            export = yaml.load(out)
        except Exception as exc:
            if "unrecognized arguments: -p" in exc_str(exc):
                lgr.warning("Could not retrieve conda environment "
                            "export from path %s: "
                            "Please use Conda 4.3.19 or greater",
                            conda_path)
            else:
                lgr.warning("Could not retrieve conda environment "
                            "export from path %s: %s", conda_path,
                            exc_str(exc))
        return export

    def _get_conda_info(self, conda_path):
        details = {}
        try:
            out, err = self._session.execute_command(
                '%s/bin/conda info --json'
                % conda_path
            )
            details = json.loads(out)
        except Exception as exc:
            lgr.warning("Could not retrieve conda info in path %s: %s",
                        conda_path, exc_str(exc))
        return details

    def _is_conda_env_path(self, path):
        return self._session.exists('%s/conda-meta' % path)

    def identify_distributions(self, paths):
        conda_paths = set()
        root_to_envs = defaultdict(list)
        # Start with all paths being set as unknown
        unknown_files = set(paths)
        # Track count of found packages and files
        found_package_count = 0
        total_file_count = len(unknown_files)

        # First, loop through all the files and identify conda paths
        for path in paths:
            conda_path = self._get_conda_env_path(path)
            if conda_path:
                if conda_path not in conda_paths:
                    conda_paths.add(conda_path)

        # Loop through conda_paths, find packages and create the
        # environments
        for idx, conda_path in enumerate(conda_paths):
            # Start with an empty channels list
            channels = []
            found_channel_names = set()

            # Find the root path for the environment
            # TODO: cache/memoize for those paths which have been considered
            # since will be asked again below
            conda_info = self._get_conda_info(conda_path)
            root_path = conda_info.get('root_prefix')
            if not root_path:
                lgr.warning("Could not find root path for conda environment %s"
                            % conda_path)
                continue
            # Retrieve the environment details
            env_export = self._get_conda_env_export(
               root_path, conda_path)
            (conda_package_details, file_to_pkg) = \
                self._get_conda_package_details(conda_path)
            (conda_pip_package_details, file_to_pip_pkg) = \
                self._get_conda_pip_package_details(env_export, conda_path)
            # Join our conda and pip packages
            conda_package_details.update(conda_pip_package_details)
            file_to_pkg.update(file_to_pip_pkg)

            # Initialize a map from packages to files that defaults to []
            pkg_to_found_files = defaultdict(list)

            # Get the conda path prefix to calculate relative paths
            path_prefix = conda_path + os.path.sep
            # Loop through unknown files, assigning them to packages if found
            for path in set(unknown_files):  # Clone the set
                if path in file_to_pkg:
                    # The file was found so remove from unknown file set
                    unknown_files.remove(path)
                    # Make relative paths if it is begins with the conda path
                    if path.startswith(path_prefix):
                        rel_path = path[len(path_prefix):]
                    else:
                        rel_path = path
                    # And add to the package
                    pkg_to_found_files[file_to_pkg[path]].append(rel_path)

            packages = []
            # Create the packages in the environment
            for package_name in conda_package_details:

                details = conda_package_details[package_name]

                # Look up or create the conda channel for the environment list
                channel_name = details.get("schannel")
                if channel_name:
                    if channel_name not in found_channel_names:
                        # New channel for our environment, so add it
                        channel_url = details.get("channel")
                        found_channel_names.add(channel_name)
                        channels.append(CondaChannel(
                            name=channel_name,
                            url=channel_url))

                location = details.get("location")
                # Create the package
                package = CondaPackage(
                    name=details.get("name"),
                    installer=details.get("installer"),
                    version=details.get("version"),
                    build=details.get("build"),
                    channel_name=channel_name,
                    size=details.get("size"),
                    md5=details.get("md5"),
                    url=details.get("url"),
                    location=location,
                    editable=details.get("editable"),
                    files=pkg_to_found_files[package_name]
                )
                packages.append(package)

                # Make editable pip packages available to other tracers.
                if location and not is_subpath(location, conda_path):
                    unknown_files.add(location)

            # Give the distribution a name
            # Determine name from path (Alt approach: use conda-env info)
            if os.path.normpath(conda_path) == os.path.normpath(root_path):
                env_name = "root"
            else:
                env_name = os.path.basename(os.path.normpath(conda_path))

            # Keep track of found package count
            found_package_count += len(packages)

            # Create the conda environment (works with root environments too)
            conda_env = CondaEnvironment(
                name=env_name,
                path=conda_path,
                packages=packages,
                channels=channels
            )
            root_to_envs[root_path].append(conda_env)

        lgr.debug(
            "%s: %d packages with %d files, and %d other files",
            self.__class__.__name__,
            found_package_count,
            total_file_count - len(unknown_files),
            len(unknown_files))

        # Find all the identified conda_roots
        conda_roots = root_to_envs.keys()
        # Loop through conda_roots and create the distributions
        for idx, root_path in enumerate(conda_roots):
            # Retrieve distribution details
            conda_info = self._get_conda_info(root_path)

            # Give the distribution a name
            if (len(conda_roots)) > 1:
                dist_name = 'conda-%d' % idx
            else:
                dist_name = 'conda'
            dist = CondaDistribution(
                name=dist_name,
                conda_version=conda_info.get("conda_version"),
                python_version=conda_info.get("python_version"),
                platform=conda_info.get("platform"),
                path=root_path,
                environments=root_to_envs[root_path]
            )
            yield dist, unknown_files



"""
    Example script to read a provenance file and then create the environment on the localhost.
    The provenance files in the niceman/examples directory are used for this example.
"""

# # Add current NeuroDebian and archived NeuroDebian repos.
# print '\nAdding the NeuroDebian repository to the server for %s %s ...' % (distribution['OS'], distribution['version'])
# # orchestrator.add_task('apt_repository', dict(repo='deb http://snapshot.debian.org/archive/debian/%s/ data main contrib non-free' % (create_date,)))
# orchestrator.add_task('apt_repository', dict(repo='deb http://neuro.debian.net/debian data main contrib non-free'))

# if distribution['version'] == '12.04':
#     # orchestrator.add_task('apt_repository', dict(repo='deb http://snapshot.debian.org/archive/debian/%s/ precise main' % (create_date,)))
#     orchestrator.add_task('apt_repository', dict(repo='deb http://neuro.debian.net/debian precise main contrib non-free'))
# else:
#     # orchestrator.add_task('apt_repository', dict(repo='deb http://snapshot.debian.org/archive/debian/%s/ trusty main' % (create_date,)))
#     orchestrator.add_task('apt_repository', dict(repo='deb http://neuro.debian.net/debian trusty main contrib non-free'))

# orchestrator.add_task('apt_key', dict(keyserver='hkp://pgp.mit.edu:80', id='0xA5D32F012649A5A9'))
# orchestrator.add_task('apt', dict(update_cache='yes'))

import logging
lgr = logging.getLogger('niceman.examples')


# Add packages to task list.
def get_url_for_packages(provenance):
    """Return url for every package (versioned) as specified in provenance

    It traverses passes provenance ...

    Examples
    --------

    >>> get_url_for_packages({'cmtk' : '3.2.2-1.4build1'})
    {'cmtk': 'http://example.com/cmtk_3.2.2-1.4build1.deb'}

    Parameters
    ----------
    provenance : TODO
      Provenance read from somewhere

    Returns
    -------
    dict
      package: url   for every package found in provenance

    """
    lgr.debug('Finding versioned urls for following provenance info: %s',
              str(provenance))
    return {
        package:  'http://example.com/%s_%s.deb' % (package, version)  # dpm.get_ubuntu_binary_pkg_url(package[0], package[1])
        for package, version in provenance.items()  # get_packages()
    }
    # orchestrator.add_task('apt', dict(name='%s=%s' % package))

# Run the orchestrator against the target host.
# orchestrator.run()



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import pytest
from .constants import NICEMAN_SPEC1_YML_FILENAME, REPROZIP_SPEC2_YML_FILENAME

NICEMAN_SPECS = [
    NICEMAN_SPEC1_YML_FILENAME,
    # This one could have been another "serialization" of the effectively
    # identical spec on which we could try to run
    # REPROZIP_SPEC1
]

REPROZIP_SPEC2 = [
    REPROZIP_SPEC2_YML_FILENAME
]


# Let's make a convenience fixture to run tests against demo1 file(s)
@pytest.fixture(params=NICEMAN_SPECS)
def demo1_spec(request):
    yield request.param


@pytest.fixture(params=REPROZIP_SPEC2)
def reprozip_spec2(request):
    yield request.param




# emacs: -*- mode: python-mode; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test functioning of the niceman main cmdline utility """

import re
import sys
from six.moves import StringIO
from mock import patch
import pytest

import niceman
from ..cmdline.main import main
from .utils import assert_equal, in_, ok_startswith


def run_main(args, exit_code=0, expect_stderr=False):
    """Run main() of the niceman, do basic checks and provide outputs

    Parameters
    ----------
    args : list
        List of string cmdline arguments to pass
    exit_code : int
        Expected exit code. Would raise AssertionError if differs
    expect_stderr : bool or string
        Either to expect stderr output. If string -- match

    Returns
    -------
    stdout, stderr  strings
       Output produced
    """
    with patch('sys.stderr', new_callable=StringIO) as cmerr:
        with patch('sys.stdout', new_callable=StringIO) as cmout:
            with pytest.raises(SystemExit) as cm:
                main(args)
            assert_equal(cm.value.code, exit_code)  # exit code must be 0
            stdout = cmout.getvalue()
            stderr = cmerr.getvalue()
            if expect_stderr == False:
                assert_equal(stderr, "")
            elif expect_stderr == True:
                # do nothing -- just return
                pass
            else:
                # must be a string
                assert_equal(stderr, expect_stderr)
    return stdout, stderr


# TODO: switch to stdout for --version output
def test_version():
    stdout, stderr = run_main(['--version'], expect_stderr=True)

    # and output should contain our version, copyright, license

    # https://hg.python.org/cpython/file/default/Doc/whatsnew/3.4.rst#l1952
    out = stdout if sys.version_info >= (3, 4) else stderr
    ok_startswith(out, 'niceman %s\n' % niceman.__version__)
    in_("Copyright", out)
    in_("Permission is hereby granted", out)


def test_help_np():
    stdout, stderr = run_main(['--help-np'])

    # Let's extract section titles:
    # enough of bin/niceman and .tox/py27/bin/niceman -- guarantee consistency! ;)
    ok_startswith(stdout, 'Usage: niceman')
    # Sections start/end with * if ran under NICEMAN_HELP2MAN mode
    sections = [l[1:-1] for l in filter(re.compile('^\*.*\*$').match, stdout.split('\n'))]
    # but order is still not guaranteed (dict somewhere)! TODO
    # see https://travis-ci.org/niceman/niceman/jobs/80519004
    # thus testing sets
    assert_equal(set(sections),
                 {'Commands for manipulating computation environments',
                  'Miscellaneous commands',
                  'General information',
                  'Global options'})

# MJT - This test incorrectly tests how the create, ls, install, etc. commands
# work as they now prompt for missing args rather than display a usage message.
# def test_usage_on_insufficient_args():
#     stdout, stderr = run_main(['create'], exit_code=1)
#     ok_startswith(stdout, 'usage:')



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from os.path import join as opj, dirname, pardir
from pytest import raises
from niceman.formats.niceman import NicemanProvenance
from niceman.distributions.debian import DebianDistribution
from niceman.distributions.conda import CondaDistribution

yaml_dir = opj(dirname(__file__), pardir, pardir, 'interface', 'tests', 'files')
multi_debian_yaml = opj(yaml_dir, 'multi_debian.yaml')
diff_1_yaml = opj(yaml_dir, 'diff_1.yaml')

def test_get_distributions():
    env = NicemanProvenance(multi_debian_yaml).get_environment()
    with raises(ValueError):
        env.get_distribution(DebianDistribution)
    dist = env.get_distribution(CondaDistribution)
    assert dist is None
    env = NicemanProvenance(diff_1_yaml).get_environment()
    dist = env.get_distribution(DebianDistribution)
    assert isinstance(dist, DebianDistribution)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import os
import tempfile

from ...cmd import Runner
# from ...distributions.singularity import SingularityDistribution
# from ...distributions.singularity import SingularityImage
from ...distributions.singularity import SingularityTracer
from ...tests.utils import skip_if_no_singularity, skip_if_no_network


@skip_if_no_network
@skip_if_no_singularity
def test_singularity_trace():

    # Download and set up singularity image file
    runner = Runner()
    runner.run(['singularity', 'pull',
        'shub://vsoch/hello-world@42e1f04ed80217895f8c960bdde6bef4d34fab59'])
    image_file = tempfile.mktemp()
    os.rename('vsoch-hello-world-master-latest.simg', image_file)

    # Test tracer class
    tracer = SingularityTracer()

    files = [image_file, 'non-existant-image']
    dist, remaining_files = next(tracer.identify_distributions(files))

    assert dist.images[0].md5 == 'ed9755a0871f04db3e14971bec56a33f'
    assert dist.images[0].bootstrap == 'docker'
    assert dist.images[0].maintainer == 'vanessasaur'
    assert dist.images[0].deffile == 'Singularity'
    assert dist.images[0].schema_version == '1.0'
    assert dist.images[0].build_date == '2017-10-15T12:52:56+00:00'
    assert dist.images[0].build_size == '333MB'
    assert dist.images[0].singularity_version == '2.4-feature-squashbuild-secbuild.g780c84d'
    assert dist.images[0].base_image == 'ubuntu:14.04'
    assert 'non-existant-image' in remaining_files

    # Clean up test files
    os.remove(image_file)

    # Run tracer against a Singularity Hub url
    files = ['shub:/GodloveD/lolcow@a59d8de3121579fe9c95ab8af0297c2e3aefd827']
    dist, remaining_files = next(tracer.identify_distributions(files))
    assert dist.images[0].md5 == 'ee4aae1ea378ad7c0299b308c703187a'
    assert dist.images[0].bootstrap == 'docker'
    assert dist.images[0].deffile == 'Singularity'
    assert dist.images[0].schema_version == '1.0'
    assert dist.images[0].build_date == '2017-10-17T19:23:53+00:00'
    assert dist.images[0].build_size == '336MB'
    assert dist.images[0].singularity_version == '2.4-feature-squashbuild-secbuild.g217367c'
    assert dist.images[0].base_image == 'ubuntu:16.04'
    assert dist.images[0].url == 'shub://GodloveD/lolcow@a59d8de3121579fe9c95ab8af0297c2e3aefd827'



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Support for Singularity distribution(s)."""

import attr
import json
import logging
import os
import tempfile
import uuid


lgr = logging.getLogger('niceman.distributions.singularity')

from .base import Package
from .base import Distribution
from .base import DistributionTracer
from .base import TypedList
from .base import _register_with_representer
from ..dochelpers import borrowdoc, exc_str
from ..utils import attrib, md5sum, chpwd


@attr.s(slots=True, frozen=True, cmp=False, hash=True)
class SingularityImage(Package):
    """Singularity image information"""
    md5 = attrib(default=attr.NOTHING)
    # Optional
    bootstrap = attrib()
    maintainer = attrib()
    deffile = attrib()
    schema_version = attrib()
    build_date = attrib()
    build_size = attrib()
    singularity_version = attrib()
    base_image = attrib()
    mirror_url = attrib()
    url = attrib()
    path = attrib()


_register_with_representer(SingularityImage)


@attr.s
class SingularityDistribution(Distribution):
    """
    Class to provide commands to Singularity.
    """

    images = TypedList(SingularityImage)

    def initiate(self, session):
        """
        Perform any initialization commands needed in the environment.

        Parameters
        ----------
        session : object
            The Session to work in
        """

        # Raise niceman.support.exceptions.CommandError exception if
        # Singularity is not to be found.
        session.execute_command(['singularity', 'selftest'])

    def install_packages(self, session):
        """
        Install the Singularity images associated to this distribution by the
        provenance into the environment.

        Parameters
        ----------
        session : object
            Session to work in
        """

        # TODO: Currently we have no way to locate the image given the metadata


_register_with_representer(SingularityDistribution)


class SingularityTracer(DistributionTracer):
    """Singularity image tracer

    If a given file is not identified as a singularity image, the files
    are quietly passed on to the next tracer.
    """

    HANDLES_DIRS = False

    @borrowdoc(DistributionTracer)
    def identify_distributions(self, files):
        if not files:
            return

        images = []
        remaining_files = set()
        url = None
        path = None

        for file_path in files:
            try:
                if file_path.startswith('shub:/'):
                    # Correct file path for path normalization in retrace.py
                    if not file_path.startswith('shub://'):
                        file_path = file_path.replace('shub:/', 'shub://')
                    temp_path = "{}.simg".format(uuid.uuid4())
                    with chpwd(tempfile.gettempdir()):
                        msg = "Downloading Singularity image {} for tracing"
                        lgr.info(msg.format(file_path))
                        self._session.execute_command(['singularity', 'pull',
                            '--name', temp_path, file_path])
                        image = json.loads(self._session.execute_command(
                            ['singularity', 'inspect', temp_path])[0])
                        url = file_path
                        md5 = md5sum(temp_path)
                        os.remove(temp_path)
                else:
                    path = os.path.abspath(file_path)
                    image = json.loads(self._session.execute_command(
                        ['singularity', 'inspect', file_path])[0])
                    md5 = md5sum(file_path)

                images.append(SingularityImage(
                    md5=md5,
                    bootstrap=image.get(
                        'org.label-schema.usage.singularity.deffile.bootstrap'),
                    maintainer=image.get('MAINTAINER'),
                    deffile=image.get(
                        'org.label-schema.usage.singularity.deffile'),
                    schema_version=image.get('org.label-schema.schema-version'),
                    build_date=image.get('org.label-schema.build-date'),
                    build_size=image.get('org.label-schema.build-size'),
                    singularity_version=image.get(
                        'org.label-schema.usage.singularity.version'),
                    base_image=image.get(
                        'org.label-schema.usage.singularity.deffile.from'),
                    mirror_url=image.get(
                        'org.label-schema.usage.singularity.deffile.mirrorurl'),
                    url=url,
                    path=path
                ))
            except Exception as exc:
                lgr.debug("Probably %s is not a Singularity image: %s",
                    file_path, exc_str(exc))
                lgr.debug(exc)
                remaining_files.add(file_path)

        if not images:
            return

        dist = SingularityDistribution(
            name="singularity",
            images=images
        )

        yield dist, remaining_files

    @borrowdoc(DistributionTracer)
    def _get_packagefields_for_files(self, files):
        return

    @borrowdoc(DistributionTracer)
    def _create_package(self, **package_fields):
        return



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import collections
import re
import six

import six.moves.builtins as __builtin__
from six.moves import shlex_quote
import time

from os.path import curdir, basename, exists, realpath, islink, join as opj, isabs, normpath, expandvars, expanduser, abspath
from six.moves.urllib.parse import quote as urlquote, unquote as urlunquote, urlsplit
from six import text_type, binary_type, PY3

import logging
import shutil
import stat
import os
import sys
import tempfile
import platform
import gc
import glob

import attr
from functools import wraps
from time import sleep
import inspect
from itertools import tee

from niceman.support.exceptions import CommandError

lgr = logging.getLogger("niceman.utils")

lgr.log(5, "Importing niceman.utils")
#
# Some useful variables
#
_platform_system = platform.system().lower()
on_windows = _platform_system == 'windows'
on_osx = _platform_system == 'darwin'
on_linux = _platform_system == 'linux'
try:
    linux_distribution = platform.linux_distribution()
    on_debian_wheezy = on_linux \
                       and linux_distribution[0] == 'debian' \
                       and linux_distribution[1].startswith('7.')
except:  # pragma: no cover
    on_debian_wheezy = False

#
# Little helpers
#

# `getargspec` has been deprecated in Python 3.
if hasattr(inspect, "getfullargspec"):
    def getargspec(func):
        """Backward-compatibility wrapper for inspect.getargspec.
        """
        # The first four elements in getfullargspec's return value match
        # getargspec's.
        return inspect.getfullargspec(func)[:4]
else:
    getargspec = inspect.getargspec


def get_func_kwargs_doc(func):
    """ Provides args for a function
    
    Parameters
    ----------
    func: str
      name of the function from which args are being requested

    Returns
    -------
    list
      of the args that a function takes in
    """
    return getargspec(func)[0]

    # TODO: format error message with descriptions of args
    # return [repr(dict(get_docstring_split(func)[1]).get(x)) for x in getargspec(func)[0]]


def assure_tuple_or_list(obj):
    """Given an object, wrap into a tuple if not list or tuple
    """
    if isinstance(obj, list) or isinstance(obj, tuple):
        return obj
    return (obj,)


def any_re_search(regexes, value):
    """Return if any of regexes (list or str) searches succesfully for value"""
    for regex in assure_tuple_or_list(regexes):
        if re.search(regex, value):
            return True
    return False


def not_supported_on_windows(msg=None):
    """A little helper to be invoked to consistently fail whenever functionality is
    not supported (yet) on Windows
    """
    if on_windows:
        raise NotImplementedError("This functionality is not yet implemented for Windows OS"
                                  + (": %s" % msg if msg else ""))


def shortened_repr(value, l=30):
    try:
        if hasattr(value, '__repr__') and (value.__repr__ is not object.__repr__):
            value_repr = repr(value)
            if not value_repr.startswith('<') and len(value_repr) > l:
                value_repr = "<<%s...>>" % (value_repr[:l-8])
            elif value_repr.startswith('<') and value_repr.endswith('>') and ' object at 0x':
                raise ValueError("I hate those useless long reprs")
        else:
            raise ValueError("gimme class")
    except Exception as e:
        value_repr = "<%s>" % value.__class__.__name__.split('.')[-1]
    return value_repr


def __auto_repr__(obj):
    attr_names = tuple()
    if hasattr(obj, '__dict__'):
        attr_names += tuple(obj.__dict__.keys())
    if hasattr(obj, '__slots__'):
        attr_names += tuple(obj.__slots__)

    items = []
    for attr in sorted(set(attr_names)):
        if attr.startswith('_'):
            continue
        value = getattr(obj, attr)
        # TODO:  should we add this feature to minimize some talktative reprs
        # such as of URL?
        #if value is None:
        #    continue
        items.append("%s=%s" % (attr, shortened_repr(value)))

    return "%s(%s)" % (obj.__class__.__name__, ', '.join(items))

def auto_repr(cls):
    """Decorator for a class to assign it an automagic quick and dirty __repr__

    It uses public class attributes to prepare repr of a class

    Original idea: http://stackoverflow.com/a/27799004/1265472
    """

    cls.__repr__ = __auto_repr__
    return cls

def is_interactive():
    """Return True if all in/outs are tty"""
    # TODO: check on windows if hasattr check would work correctly and add value:
    #
    return sys.stdin.isatty() and sys.stdout.isatty() and sys.stderr.isatty()


import hashlib


def md5sum(filename):
    with open(filename, 'rb') as f:
        return hashlib.md5(f.read()).hexdigest()


def sorted_files(dout):
    """Return a (sorted) list of files under dout
    """
    return sorted(sum([[opj(r, f)[len(dout)+1:] for f in files]
                       for r,d,files in os.walk(dout)
                       if not '.git' in r], []))

from os.path import sep as dirsep
_VCS_REGEX = '%s\.(?:git|gitattributes|svn|bzr|hg)(?:%s|$)' % (dirsep, dirsep)
_NICEMAN_REGEX = '%s\.(?:niceman)(?:%s|$)' % (dirsep, dirsep)


def find_files(regex, topdir=curdir, exclude=None, exclude_vcs=True, exclude_niceman=False, dirs=False):
    """Generator to find files matching regex

    Parameters
    ----------
    regex: basestring
    exclude: basestring, optional
      Matches to exclude
    exclude_vcs:
      If True, excludes commonly known VCS subdirectories.  If string, used
      as regex to exclude those files (regex: `%r`)
    exclude_niceman:
      If True, excludes files known to be niceman meta-data files (e.g. under
      .niceman/ subdirectory) (regex: `%r`)
    topdir: basestring, optional
      Directory where to search
    dirs: bool, optional
      Either to match directories as well as files
    """

    for dirpath, dirnames, filenames in os.walk(topdir):
        names = (dirnames + filenames) if dirs else filenames
        # TODO: might want to uniformize on windows to use '/'
        paths = (opj(dirpath, name) for name in names)
        for path in filter(re.compile(regex).search, paths):
            path = path.rstrip(dirsep)
            if exclude and re.search(exclude, path):
                continue
            if exclude_vcs and re.search(_VCS_REGEX, path):
                continue
            if exclude_niceman and re.search(_NICEMAN_REGEX, path):
                continue
            yield path
find_files.__doc__ %= (_VCS_REGEX, _NICEMAN_REGEX)


def expandpath(path, force_absolute=True):
    """Expand all variables and user handles in a path.

    By default return an absolute path
    """
    path = expandvars(expanduser(path))
    if force_absolute:
        path = abspath(path)
    return path


def is_explicit_path(path):
    """Return whether a path explicitly points to a location

    Any absolute path, or relative path starting with either '../' or
    './' is assumed to indicate a location on the filesystem. Any other
    path format is not considered explicit."""
    path = expandpath(path, force_absolute=False)
    return isabs(path) \
        or path.startswith(os.curdir + os.sep) \
        or path.startswith(os.pardir + os.sep)

def rotree(path, ro=True, chmod_files=True):
    """To make tree read-only or writable

    Parameters
    ----------
    path : string
      Path to the tree/directory to chmod
    ro : bool, optional
      Either to make it R/O (default) or RW
    chmod_files : bool, optional
      Either to operate also on files (not just directories)
    """
    if ro:
        chmod = lambda f: os.chmod(f, os.stat(f).st_mode & ~stat.S_IWRITE)
    else:
        chmod = lambda f: os.chmod(f, os.stat(f).st_mode | stat.S_IWRITE | stat.S_IREAD)

    for root, dirs, files in os.walk(path, followlinks=False):
        if chmod_files:
            for f in files:
                fullf = opj(root, f)
                # might be the "broken" symlink which would fail to stat etc
                if exists(fullf):
                    chmod(fullf)
        chmod(root)


def rmtree(path, chmod_files='auto', *args, **kwargs):
    """To remove git-annex .git it is needed to make all files and directories writable again first

    Parameters
    ----------
    chmod_files : string or bool, optional
       Either to make files writable also before removal.  Usually it is just
       a matter of directories to have write permissions.
       If 'auto' it would chmod files on windows by default
    `*args` :
    `**kwargs` :
       Passed into shutil.rmtree call
    """
    # Give W permissions back only to directories, no need to bother with files
    if chmod_files == 'auto':
        chmod_files = on_windows

    if not os.path.islink(path):
        rotree(path, ro=False, chmod_files=chmod_files)
        shutil.rmtree(path, *args, **kwargs)
    else:
        # just remove the symlink
        os.unlink(path)


def rmtemp(f, *args, **kwargs):
    """Wrapper to centralize removing of temp files so we could keep them around

    It will not remove the temporary file/directory if NICEMAN_TESTS_KEEPTEMP
    environment variable is defined
    """
    if not os.environ.get('NICEMAN_TESTS_KEEPTEMP'):
        if not os.path.lexists(f):
            lgr.debug("Path %s does not exist, so can't be removed" % f)
            return
        lgr.log(5, "Removing temp file: %s" % f)
        # Can also be a directory
        if os.path.isdir(f):
            rmtree(f, *args, **kwargs)
        else:
            for i in range(10):
                try:
                    os.unlink(f)
                except OSError as e:
                    if i < 9:
                        sleep(0.1)
                        continue
                    else:
                        raise
                break
    else:
        lgr.info("Keeping temp file: %s" % f)


def file_basename(name, return_ext=False):
    """
    Strips up to 2 extensions of length up to 4 characters and starting with alpha
    not a digit, so we could get rid of .tar.gz etc
    """
    bname = basename(name)
    fbname = re.sub('(\.[a-zA-Z_]\S{1,4}){0,2}$', '', bname)
    if return_ext:
        return fbname, bname[len(fbname)+1:]
    else:
        return fbname


def escape_filename(filename):
    """Surround filename in "" and escape " in the filename
    """
    filename = filename.replace('"', r'\"').replace('`', r'\`')
    filename = '"%s"' % filename
    return filename


def encode_filename(filename):
    """Encode unicode filename
    """
    if isinstance(filename, text_type):
        return filename.encode(sys.getfilesystemencoding())
    else:
        return filename

if on_windows:
    def lmtime(filepath, mtime):
        """Set mtime for files.  On Windows a merely adapter to os.utime
        """
        os.utime(filepath, (time.time(), mtime))
else:
    def lmtime(filepath, mtime):
        """Set mtime for files, while not de-referencing symlinks.

        To overcome absence of os.lutime

        Works only on linux and OSX ATM
        """
        from .cmd import Runner
        # convert mtime to format touch understands [[CC]YY]MMDDhhmm[.SS]
        smtime = time.strftime("%Y%m%d%H%M.%S", time.localtime(mtime))
        lgr.log(3, "Setting mtime for %s to %s == %s", filepath, mtime, smtime)
        Runner().run(['touch', '-h', '-t', '%s' % smtime, filepath])
        rfilepath = realpath(filepath)
        if islink(filepath) and exists(rfilepath):
            # trust noone - adjust also of the target file
            # since it seemed like downloading under OSX (was it using curl?)
            # didn't bother with timestamps
            lgr.log(3, "File is a symlink to %s Setting mtime for it to %s",
                    rfilepath, mtime)
            os.utime(rfilepath, (time.time(), mtime))
        # doesn't work on OSX
        # Runner().run(['touch', '-h', '-d', '@%s' % mtime, filepath])


def assure_list(s):
    """Given not a list, would place it into a list. If None - empty list is returned

    Parameters
    ----------
    s: list or anything
    """

    if isinstance(s, list):
        return s
    elif s is None:
        return []
    else:
        return [s]


def assure_list_from_str(s, sep='\n'):
    """Given a multiline string convert it to a list of return None if empty

    Parameters
    ----------
    s: str or list
    """

    if not s:
        return None

    if isinstance(s, list):
        return s
    return s.split(sep)


def assure_dict_from_str(s, **kwargs):
    """Given a multiline string with key=value items convert it to a dictionary

    Parameters
    ----------
    s: str or dict

    Returns None if input s is empty
    """

    if not s:
        return None

    if isinstance(s, dict):
        return s

    out = {}
    for value_str in assure_list_from_str(s, **kwargs):
        if '=' not in value_str:
            raise ValueError("{} is not in key=value format".format(repr(value_str)))
        k, v = value_str.split('=', 1)
        if k in out:
            err  = "key {} was already defined in {}, but new value {} was provided".format(k, out, v)
            raise ValueError(err)
        out[k] = v
    return out

def only_with_values(d):
    """Given a dictionary, return the one only with entries which had non-null values"""
    # to maintain OrderedDict do explicit d.__class__
    return d.__class__((k, v) for k,v in d.items() if v)

def assure_bytes(s, encoding='utf-8'):
    """Convert/encode unicode to str (PY2) or bytes (PY3) if of 'text_type'

    Parameters
    ----------
    encoding: str, optional
      Encoding to use.  "utf-8" is the default
    """
    if not isinstance(s, text_type):
        return s
    return s.encode(encoding)


def assure_unicode(s, encoding=None, confidence=None):
    """Convert/decode to unicode (PY2) or str (PY3) if of 'binary_type'

    Parameters
    ----------
    encoding: str, optional
      Encoding to use.  If None, "utf-8" is tried, and then if not a valid
      UTF-8, encoding will be guessed
    confidence: float, optional
      A value between 0 and 1, so if guessing of encoding is of lower than
      specified confidence, ValueError is raised
    """
    if not isinstance(s, binary_type):
        return s
    if encoding is None:
        # Figure out encoding, defaulting to 'utf-8' which is our common
        # target in contemporary digital society
        try:
            return s.decode('utf-8')
        except UnicodeDecodeError as exc:
            from .dochelpers import exc_str
            lgr.debug("Failed to decode a string as utf-8: %s", exc_str(exc))
        # And now we could try to guess
        from chardet import detect
        enc = detect(s)
        denc = enc.get('encoding', None)
        if denc:
            denc_confidence = enc.get('confidence', 0)
            if confidence is not None and  denc_confidence < confidence:
                raise ValueError(
                    "Failed to auto-detect encoding with high enough "
                    "confidence. Highest confidence was %s for %s"
                    % (denc_confidence, denc)
                )
            return s.decode(denc)
        else:
            raise ValueError(
                "Could not decode value as utf-8, or to guess its encoding: %s"
                % repr(s)
            )
    else:
        return s.decode(encoding)


def unique(seq, key=None):
    """Given a sequence return a list only with unique elements while maintaining order

    This is the fastest solution.  See
    https://www.peterbe.com/plog/uniqifiers-benchmark
    and
    http://stackoverflow.com/a/480227/1265472
    for more information.
    Enhancement -- added ability to compare for uniqueness using a key function

    Parameters
    ----------
    seq:
      Sequence to analyze
    key: callable, optional
      Function to call on each element so we could decide not on a full
      element, but on its member etc
    """
    seen = set()
    seen_add = seen.add
    if not key:
        return [x for x in seq if not (x in seen or seen_add(x))]
    else:
        # OPT: could be optimized, since key is called twice, but for our cases
        # should be just as fine
        return [x for x in seq if not (key(x) in seen or seen_add(key(x)))]


def partition(items, predicate=bool):
    """Partition `items` by `predicate`.

    Parameters
    ----------
    items : iterable
    predicate : callable
        A function that will be mapped over each element in `items`. The
        elements will partitioned based on whether the return value is false or
        true.

    Returns
    -------
    A tuple with two generators, the first for 'false' items and the second for
    'true' ones.

    Notes
    -----
    Taken from Peter Otten's snippet posted at
    https://nedbatchelder.com/blog/201306/filter_a_list_into_two_parts.html
    """
    a, b = tee((predicate(item), item) for item in items)
    return ((item for pred, item in a if not pred),
            (item for pred, item in b if pred))


#
# Decorators
#

# Borrowed from pandas
# Copyright: 2011-2014, Lambda Foundry, Inc. and PyData Development Team
# Licese: BSD-3
def optional_args(decorator):
    """allows a decorator to take optional positional and keyword arguments.
        Assumes that taking a single, callable, positional argument means that
        it is decorating a function, i.e. something like this::

            @my_decorator
            def function(): pass

        Calls decorator with decorator(f, `*args`, `**kwargs`)"""

    @wraps(decorator)
    def wrapper(*args, **kwargs):
        def dec(f):
            return decorator(f, *args, **kwargs)

        is_decorating = not kwargs and len(args) == 1 and isinstance(args[0], collections.Callable)
        if is_decorating:
            f = args[0]
            args = []
            return dec(f)
        else:
            return dec

    return wrapper


# TODO: just provide decorators for tempfile.mk* functions. This is ugly!
def get_tempfile_kwargs(tkwargs={}, prefix="", wrapped=None):
    """Updates kwargs to be passed to tempfile. calls depending on env vars
    """
    # operate on a copy of tkwargs to avoid any side-effects
    tkwargs_ = tkwargs.copy()

    # TODO: don't remember why I had this one originally
    # if len(targs)<2 and \
    if not 'prefix' in tkwargs_:
        tkwargs_['prefix'] = '_'.join(
            ['niceman_temp'] +
            ([prefix] if prefix else []) +
            ([''] if (on_windows or not wrapped)
                  else [wrapped.__name__]))

    directory = os.environ.get('NICEMAN_TESTS_TEMPDIR')
    if directory and 'dir' not in tkwargs_:
        tkwargs_['dir'] = directory

    return tkwargs_

@optional_args
def line_profile(func):
    """Q&D helper to line profile the function and spit out stats
    """
    import line_profiler
    prof = line_profiler.LineProfiler()

    @wraps(func)
    def newfunc(*args, **kwargs):
        try:
            pfunc = prof(func)
            return pfunc(*args, **kwargs)
        finally:
            prof.print_stats()
    return newfunc

#
# Context Managers
#

from contextlib import contextmanager

@contextmanager
def swallow_outputs():
    """Context manager to help consuming both stdout and stderr, and print()

    stdout is available as cm.out and stderr as cm.err whenever cm is the
    yielded context manager.
    Internally uses temporary files to guarantee absent side-effects of swallowing
    into StringIO which lacks .fileno.

    print mocking is necessary for some uses where sys.stdout was already bound
    to original sys.stdout, thus mocking it later had no effect. Overriding
    print function had desired effect
    """

    debugout = sys.stdout
    class StringIOAdapter(object):
        """Little adapter to help getting out/err values
        """
        def __init__(self):
            kw = get_tempfile_kwargs({}, prefix="outputs")

            self._out = open(tempfile.mktemp(**kw), 'w')
            self._err = open(tempfile.mktemp(**kw), 'w')

        def _read(self, h):
            with open(h.name) as f:
                return f.read()

        @property
        def out(self):
            self._out.flush()
            return self._read(self._out)

        @property
        def err(self):
            self._err.flush()
            return self._read(self._err)

        @property
        def handles(self):
            return self._out, self._err

        def cleanup(self):
            self._out.close()
            self._err.close()
            out_name = self._out.name
            err_name = self._err.name
            del self._out
            del self._err
            gc.collect()
            rmtemp(out_name)
            rmtemp(err_name)



    def fake_print(*args, **kwargs):
        sep = kwargs.pop('sep', ' ')
        end = kwargs.pop('end', '\n')
        file = kwargs.pop('file', sys.stdout)

        if file in (oldout, olderr, sys.stdout, sys.stderr):
            # we mock
            sys.stdout.write(sep.join(args) + end)
        else:
            # must be some other file one -- leave it alone
            oldprint(*args, sep=sep, end=end, file=file)

    from .ui import ui
    # preserve -- they could have been mocked already
    oldprint = getattr(__builtin__, 'print')
    oldout, olderr = sys.stdout, sys.stderr
    olduiout = ui.out
    adapter = StringIOAdapter()

    try:
        sys.stdout, sys.stderr = adapter.handles
        ui.out = adapter.handles[0]
        setattr(__builtin__, 'print', fake_print)

        yield adapter
    finally:
        sys.stdout, sys.stderr, ui.out = oldout, olderr, olduiout
        setattr(__builtin__, 'print',  oldprint)
        adapter.cleanup()


@contextmanager
def swallow_logs(new_level=None):
    """Context manager to consume all logs.

    """
    lgr = logging.getLogger("niceman")

    # Keep old settings
    old_level = lgr.level
    old_handlers = lgr.handlers

    # Let's log everything into a string
    # TODO: generalize with the one for swallow_outputs
    class StringIOAdapter(object):
        """Little adapter to help getting out values

        And to stay consistent with how swallow_outputs behaves
        """
        def __init__(self):
            kw = dict()
            get_tempfile_kwargs(kw, prefix="logs")

            self._out = open(tempfile.mktemp(**kw), 'w')

        def _read(self, h):
            with open(h.name) as f:
                return f.read()

        @property
        def out(self):
            self._out.flush()
            return self._read(self._out)

        @property
        def lines(self):
            return self.out.split('\n')

        @property
        def handle(self):
            return self._out

        def cleanup(self):
            self._out.close()
            out_name = self._out.name
            del self._out
            gc.collect()
            rmtemp(out_name)

    adapter = StringIOAdapter()
    lgr.handlers = [logging.StreamHandler(adapter.handle)]
    if old_level < logging.DEBUG:  # so if HEAVYDEBUG etc -- show them!
        lgr.handlers += old_handlers
    if isinstance(new_level, str):
        new_level = getattr(logging, new_level)

    if new_level is not None:
        lgr.setLevel(new_level)

    try:
        yield adapter
    finally:
        lgr.handlers, lgr.level = old_handlers, old_level
        adapter.cleanup()


#
# Additional handlers
#
_sys_excepthook = sys.excepthook  # Just in case we ever need original one
def setup_exceptionhook(ipython=False):
    """Overloads default sys.excepthook with our exceptionhook handler.

       If interactive, our exceptionhook handler will invoke
       pdb.post_mortem; if not interactive, then invokes default handler.
    """

    def _niceman_pdb_excepthook(type, value, tb):
        import traceback
        traceback.print_exception(type, value, tb)
        print()
        if is_interactive():
            import pdb
            pdb.post_mortem(tb)

    if ipython:
        from IPython.core import ultratb
        sys.excepthook = ultratb.FormattedTB(mode='Verbose',
                                             # color_scheme='Linux',
                                             call_pdb=is_interactive())
    else:
        sys.excepthook = _niceman_pdb_excepthook


def assure_dir(*args):
    """Make sure directory exists.

    Joins the list of arguments to an os-specific path to the desired
    directory and creates it, if it not exists yet.
    """
    dirname = opj(*args)
    if not exists(dirname):
        os.makedirs(dirname)
    return dirname

def updated(d, update):
    """Return a copy of the input with the 'update'

    Primarily for updating dictionaries
    """
    d = d.copy()
    d.update(update)
    return d

def getpwd():
    """Try to return a CWD without dereferencing possible symlinks

    If no PWD found in the env, output of getcwd() is returned
    """
    try:
        return os.environ['PWD']
    except KeyError:
        return os.getcwd()

class chpwd(object):
    """Wrapper around os.chdir which also adjusts environ['PWD']

    The reason is that otherwise PWD is simply inherited from the shell
    and we have no ability to assess directory path without dereferencing
    symlinks.

    If used as a context manager it allows to temporarily change directory
    to the given path
    """
    def __init__(self, path, mkdir=False, logsuffix=''):

        if path:
            pwd = getpwd()
            self._prev_pwd = pwd
        else:
            self._prev_pwd = None
            return

        if not isabs(path):
            path = normpath(opj(pwd, path))
        if not os.path.exists(path) and mkdir:
            self._mkdir = True
            os.mkdir(path)
        else:
            self._mkdir = False
        lgr.debug("chdir %r -> %r %s", self._prev_pwd, path, logsuffix)
        os.chdir(path)  # for grep people -- ok, to chdir here!
        os.environ['PWD'] = path

    def __enter__(self):
        # nothing more to do really, chdir was in the constructor
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._prev_pwd:
            # Need to use self.__class__ so this instance, if the entire
            # thing mocked during the test, still would use correct chpwd
            self.__class__(self._prev_pwd, logsuffix="(coming back)")


def knows_annex(path):
    """Returns whether at a given path there is information about an annex

    It is just a thin wrapper around GitRepo.is_with_annex() classmethod
    which also checks for `path` to exist first.

    This includes actually present annexes, but also uninitialized ones, or
    even the presence of a remote annex branch.
    """
    from os.path import exists
    if not exists(path):
        lgr.debug("No annex: test path {0} doesn't exist".format(path))
        return False
    from niceman.support.gitrepo import GitRepo
    return GitRepo(path, init=False, create=False).is_with_annex()


@contextmanager
def make_tempfile(content=None, wrapped=None, **tkwargs):
    """Helper class to provide a temporary file name and remove it at the end (context manager)

    Parameters
    ----------
    mkdir : bool, optional (default: False)
        If True, temporary directory created using tempfile.mkdtemp()
    content : str or bytes, optional
        Content to be stored in the file created
    wrapped : function, optional
        If set, function name used to prefix temporary file name
    `**tkwargs`:
        All other arguments are passed into the call to tempfile.mk{,d}temp(),
        and resultant temporary filename is passed as the first argument into
        the function t.  If no 'prefix' argument is provided, it will be
        constructed using module and function names ('.' replaced with
        '_').

    To change the used directory without providing keyword argument 'dir' set
    NICEMAN_TESTS_TEMPDIR.

    Examples
    --------
        >>> from os.path import exists
        >>> from niceman.utils import make_tempfile
        >>> with make_tempfile() as fname:
        ...    k = open(fname, 'w').write('silly test')
        >>> assert not exists(fname)  # was removed

        >>> with make_tempfile(content="blah") as fname:
        ...    assert open(fname).read() == "blah"
    """

    if tkwargs.get('mkdir', None) and content is not None:
        raise ValueError("mkdir=True while providing content makes no sense")

    tkwargs_ = get_tempfile_kwargs(tkwargs, wrapped=wrapped)

    # if NICEMAN_TESTS_TEMPDIR is set, use that as directory,
    # let mktemp handle it otherwise. However, an explicitly provided
    # dir=... will override this.
    mkdir = tkwargs_.pop('mkdir', False)

    filename = {False: tempfile.mktemp,
                True: tempfile.mkdtemp}[mkdir](**tkwargs_)
    filename = realpath(filename)

    if content:
        with open(filename, 'w' + ('b' if isinstance(content, binary_type) else '')) as f:
            f.write(content)

    if __debug__:
        # TODO mkdir
        lgr.debug('Created temporary thing named %s"' % filename)
    try:
        yield filename
    finally:
        # glob here for all files with the same name (-suffix)
        # would be useful whenever we requested .img filename,
        # and function creates .hdr as well
        lsuffix = len(tkwargs_.get('suffix', ''))
        filename_ = lsuffix and filename[:-lsuffix] or filename
        filenames = glob.glob(filename_ + '*')
        if len(filename_) < 3 or len(filenames) > 5:
            # For paranoid yoh who stepped into this already ones ;-)
            lgr.warning("It is unlikely that it was intended to remove all"
                        " files matching %r. Skipping" % filename_)
            return
        for f in filenames:
            try:
                rmtemp(f)
            except OSError:
                pass


def _path_(p):
    """Given a path in POSIX" notation, regenerate one in native to the env one"""
    if on_windows:
        return opj(p.split('/'))
    else:
        # Assume that all others as POSIX compliant so nothing to be done
        return p


def is_unicode(s):
    """Return true if an object is unicode"""
    return isinstance(s, six.text_type)


def is_binarystring(s):
    """Return true if an object is a binary string (not unicode)"""
    return isinstance(s, six.binary_type)


def to_unicode(s, encoding="utf-8"):
    """Converts any type string to unicode"""
    if is_unicode(s):
        return s
    else:
        return s.decode(encoding=encoding)


def to_binarystring(s, encoding="utf-8"):
    """Converts any type string to binarystring"""
    if is_binarystring(s):
        return s
    else:
        return s.encode(encoding=encoding)


def safe_write(ostream, s, encoding="utf-8"):
    """Safely write different string types to an output stream"""
    try:  # Try unicode, and upon failure try binary_string
        ostream.write(to_unicode(s, encoding))
    except (TypeError, UnicodeEncodeError):
        ostream.write(to_binarystring(s, encoding))


def generate_unique_name(pattern, nameset):
    """Create a unique numbered name from a pattern and a set

    Parameters
    ----------
    pattern: basestring
      The pattern for the name (to be used with %) that includes one %d
      location
    nameset: collection
      Collection (set or list) of existing names. If the generated name is
      used, then add the name to the nameset.

    Returns
    -------
    str
      The generated unique name
    """
    i = 0
    while True:
        n = pattern % i
        i += 1
        if n not in nameset:
            return n


# http://stackoverflow.com/questions/1151658/python-hashable-dicts
class HashableDict(dict):
    """Dict that can be used as keys"""
    def __hash__(self):
        return hash(frozenset(self.values()))


def get_cmd_batch_len(arg_list, cmd_len):
    """Estimate the maximum batch length for a given argument list

    To make sure we don't call shell commands with too many arguments
    this function looks at an argument list and the command length without
    any arguments, and estimates the number of arguments we want to batch
    together at one time.

    Parameters
    ----------
    arg_list : list
      The list to process in the command
    cmd_len : number
      The length of the command without arguments

    Returns
    -------
    number
      The maximum number in a single batch
    """
    if not arg_list:
        raise ValueError("Cannot batch an empty argument list")
    # Pick a conservative max command-line length
    try:
        _MAX_LEN_CMDLINE = os.sysconf(str("SC_ARG_MAX")) // 2
    except (ValueError, AttributeError):
        _MAX_LEN_CMDLINE = 2048
    # Find out how many files we can query at once
    max_len = max(map(len, arg_list))
    return max((_MAX_LEN_CMDLINE - cmd_len) // (max_len + 1), 1)


def join_sequence_of_dicts(seq):
    """
    Joins a sequence of dicts into a single dict

    Parameters
    ----------
    seq: sequence
        Sequence of dicts to join

    Returns
    -------
    dict

    Raises
    ------
    RuntimeError if a duplicate key is encountered.
    """
    r = {}
    for d in seq:
        for k, v in d.items():
            if k in r:
                raise RuntimeError("Duplicate key %r (new value: %r, "
                                   "was: %r)" % (k, v, r[k]))
            r[k] = v
    return r


def cmd_err_filter(err_string):
    """
    Creates a filter for CommandErrors that match a specific error string

    Parameters
    ----------
    err_string: basestring
        The error string we want to match

    Returns
    -------
    func object -> boolean
    """
    return (lambda x: isinstance(x, CommandError) and
            err_string in to_unicode(x.stderr, "utf-8"))


def execute_command_batch(session, command, args, exception_filter=None):
    """
    Generator that executes session.execute_command, with batches of args

    We want to call commands like "apt-cache policy" on a large number of
    packages, but risk creating command-lines that are too long. This
    function is a generator that will call execute_command but with batches
    of arguments (to stay within the command-line length limit) and yield the
    results.

    Parameters
    ----------
    session
      Session object that implements the execute_command() member
    command : sequence
      The command that we wish to execute
    args : sequence
      The long list of additional arguments we wish to pass to the command
    exception_filter : func x -> bool
      A filter of exception types that the calling code will gracefully handle

    Returns
    -------
    (out, err, exception)
        stdout of the command, stderr of the command, and an exception
        that is in the list of expected exceptions

    """
    cmd_length = sum(map(len, command)) + len(command)
    num_args = get_cmd_batch_len(args, cmd_length)
    args = list(args)  # we might get in with a set
    while args:
        batch, args = args[:num_args], args[num_args:]
        try:
            out, err = session.execute_command(
                command + batch
            )
            out = to_unicode(out, "utf-8")
            yield (out, err, None)
        except Exception as e:
            if exception_filter and exception_filter(e):
                yield (None, None, e)
            else:
                raise


def items_to_dict(l, attrs='name', ordered=False):
    """Given a list of attr instances, return a dict using specified attrs as keys
    
    Parameters
    ----------
    attrs : str or list of str
      Which attributes of the items to use to group
    ordered : bool, optional
      Either to return an ordered dictionary following the original order of items in the list
    
    Raises
    ------
    ValueError
        If there is a conflict - multiple items with the same attrs used for key
    
    Returns
    -------
    dict or collections.OrderedDict
    """
    many = isinstance(attrs, (list, tuple))
    out = (collections.OrderedDict if ordered else dict)()
    for i in l:
        k = tuple(getattr(i, a) for a in attrs) if many else getattr(i, attrs)
        if k in out:
            raise ValueError(
                "We already saw entry for %s: %s.  Not adding %s",
                k, out[k], i
            )
        out[k] = i
    return out


# TODO: just absorb into SpecObject __init__ but would require more handling
# to allow *args as well

def instantiate_attr_object(item_type, items):
    """Instantiate item_type given items (for a list or dict)

    Provides a more informative exception message in case if some arguments
    are incorrect
    """
    try:
        if issubclass(item_type, list):
            return item_type(items)
        else:
            return item_type(**items)
    except TypeError as exc:
        if "unexpected keyword" in str(exc):
            known_kws = [i.name for i in item_type.__attrs_attrs__]
            incorrect_kws = set(items.keys()).difference(known_kws)
            if incorrect_kws:
                # Provide a more informative message
                raise TypeError(
                    "Following provided arguments are not known to %s: %s.  "
                    "Known but not yet provided are: %s"
                    % (item_type.__name__,
                       ', '.join(incorrect_kws),
                       ', '.join(sorted(set(known_kws).difference(items))))
                )
        # if couldn't figure it out -- just raise original
        raise


def attrib(*args, **kwargs):
    """Extend the attr.ib to include our metadata elements.

    ATM we support additional keyword args which are then stored within
    `metadata`:
    - `doc` for documentation to describe the attribute (e.g. in --help)

    Also, when the `default` argument of attr.ib is unspecified, set it to
    None.
    """
    doc = kwargs.pop('doc', None)
    metadata = kwargs.get('metadata', {})
    if doc:
        metadata['doc'] = doc
    if metadata:
        kwargs['metadata'] = metadata
    return attr.ib(*args, default=kwargs.pop('default', None), **kwargs)


class PathRoot(object):
    """Find the root of paths based on a predicate function.

    The path -> root mapping is cached across calls.

    Parameters
    ----------
    predicate : callable
        A callable that will be passed a path and should return true
        if that path should be considered a root.
    """
    def __init__(self, predicate):
        self._pred = predicate
        self._cache = {}  # path -> root

    def __call__(self, path):
        """Find root of `path` based on `predicate`.

        Parameters
        ----------
        path : str
            Find this path's root.

        Returns
        -------
        str or None
        """
        to_cache = []
        root = None
        for pth in self._walk_up(path):
            if pth in self._cache:
                root = self._cache[pth]
                break

            to_cache.append(pth)

            if self._pred(pth):
                root = pth
                break

        for pth in to_cache:
            self._cache[pth] = root
        return root

    @staticmethod
    def _walk_up(path):
        """Yield PATH, chopping off the right-most directory each iteration.

        Parameters
        ----------
        path : string
        """
        while path not in [os.path.pathsep, os.path.sep, ""]:
            yield path
            path = os.path.dirname(path)


def is_subpath(path, directory):
    """Test whether `path` is below (or is itself) `directory`.

    Symbolic links are not resolved before the check.
    """
    return not os.path.relpath(path, directory).startswith(os.path.pardir)


SemanticVersion = collections.namedtuple("SemanticVersion",
                                         ["major", "minor", "patch", "tag"])


def parse_semantic_version(version):
    """Split version into major, minor, patch, and tag components.

    Parameters
    ----------
    version : str
        A version string X.Y.Z.  X, Y, and Z must be digits.  Any remaining
        text is treated as a tag (e.g., "-rc1").

    Returns
    -------
    A namedtuple with the form (major, minor, patch, tag).
    """
    m = re.match(r"(?P<major>[0-9]+)\.(?P<minor>[0-9]+)\.(?P<patch>[0-9]+)"
                 r"(?P<tag>.*)",
                 version)
    if m:
        return SemanticVersion(*m.groups())
    else:
        raise ValueError(
            "{} does not appear to follow semantic versioning".format(version))


def command_as_string(command):
    """Convert `command` to the string representation.

    Parameters
    ----------
    command : list or str
        If it is a list, convert it to a string, quoting each element as
        needed.  If it is a string, it is returned as is.
    """
    if isinstance(command, list):
        command = " ".join(map(shlex_quote, command))
    return command


lgr.log(5, "Done importing niceman.utils")



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Helper utility to delete an environment
"""

__docformat__ = 'restructuredtext'

import re

from .base import Interface
import niceman.interface.base # Needed for test patching
from .common_opts import resref_arg
from .common_opts import resref_type_opt
from ..support.param import Parameter
from ..support.constraints import EnsureStr
from ..resource import get_manager

from logging import getLogger
lgr = getLogger('niceman.api.delete')


class Delete(Interface):
    """Delete a computation environment

    Examples
    --------

      $ niceman delete my-resource

    """

    _params_ = dict(
        resref=resref_arg,
        # XXX reenable when we support working with multiple instances at once
        # resource_type=Parameter(
        #     args=("-t", "--resource-type"),
        #     doc="""Resource type to work on""",
        #     constraints=EnsureStr(),
        # ),
        resref_type=resref_type_opt,
        skip_confirmation=Parameter(
            args=("--skip-confirmation",),
            action="store_true",
            doc="Delete resource without prompting user for confirmation",
        ),
    )

    @staticmethod
    def __call__(resref, resref_type="auto", skip_confirmation=False):
        from niceman.ui import ui

        manager = get_manager()
        resource = manager.get_resource(resref, resref_type)

        if skip_confirmation:
            response = True
        else:
            response = ui.yesno(
                "Delete the resource '{}'? (ID: {})".format(
                    resource.name, resource.id[:20]),
                default="no"
            )

        if response:
            resource.connect()
            manager.delete(resource)
            lgr.info("Deleted the environment %s", resource.name)



#!/usr/bin/env python
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NICEMAN package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import platform

from os.path import sep as pathsep
from os.path import join as opj
from os.path import splitext
from os.path import dirname

from setuptools import findall
from setuptools import setup, find_packages

# manpage build imports
from setup_support import BuildManPage
from setup_support import BuildRSTExamplesFromScripts
# from setup_support import BuildConfigInfo
from setup_support import get_version


def findsome(subdir, extensions):
    """Find files under subdir having specified extensions

    Leading directory (niceman) gets stripped
    """
    return [
        f.split(pathsep, 1)[1] for f in findall(opj('niceman', subdir))
        if splitext(f)[-1].lstrip('.') in extensions
    ]

# niceman version to be installed
version = get_version()

# Only recentish versions of find_packages support include
# niceman_pkgs = find_packages('.', include=['niceman*'])
# so we will filter manually for maximal compatibility
niceman_pkgs = [pkg for pkg in find_packages('.') if pkg.startswith('niceman')]

requires = {
    'core': [
        'appdirs',
        'attrs>=16.3.0',
        'humanize',
        'mock',  # mock is also used for auto.py, not only for testing
        'pyyaml',
        'six>=1.9',
        'tqdm',
        'fabric>=2.3.1',
        'cryptography>=1.5',
        'pytz',
        'scp',
        'pycrypto',
        'pyOpenSSL==16.2.0',
        'requests',
        'reprozip; sys_platform=="linux" or sys_platform=="linux2"',
        'rpaths',
    ],
    'debian': [
        'python-debian',
        # on Debian systems will come via apt
        #'python-apt>=0.8.9.1',
        # Origin.codename was introduced in 0.8.9
        # unfortunately not on pypi but needed by python-apt
        #'http://archive.ubuntu.com/ubuntu/pool/main/p/python-apt/python-apt_0.9.3.5.tar.xz'
        'chardet',  # python-debian misses dependency on it
    ],
    'docker': [
        'docker-py>=0.3.2',
        'dockerpty',
    ],
    'aws': [
        'boto3',
    ],
    'meta': [
        'rdflib',
    ],
    'tests': [
        'mock',
        'pytest>=3.3.0',
    ]
}

requires['full'] = sum(list(requires.values()), [])

# Now add additional ones useful for development
requires.update({
    'devel-docs': [
        # used for converting README.md -> .rst for long_description
        #  TODO: enable whenever we are ready! otherwise requires installation
        #        of pandoc first via apt-get
        # 'pypandoc',
        # Documentation
        'sphinx',
        'sphinx-rtd-theme',
    ],
    'devel-utils': [
        'line-profiler',
        # necessary for accessing SecretStorage keyring (system wide Gnome
        # keyring)  but not installable on travis, IIRC since it needs connectivity
        # to the dbus whenever installed or smth like that, thus disabled here
        # but you might need it
        # 'dbus-python',
    ],
    'devel-neuroimaging': [
        # Specifically needed for tests here (e.g. example scripts testing)
        'nibabel',
    ]
})
requires['devel'] = sum(list(requires.values()), [])


# let's not build manpages and examples automatically (gh-896)
# configure additional command for custom build steps
#class DataladBuild(build_py):
#    def run(self):
#        self.run_command('build_manpage')
#        self.run_command('build_examples')
#        build_py.run(self)

cmdclass = {
    'build_manpage': BuildManPage,
    # 'build_examples': BuildRSTExamplesFromScripts,
    # 'build_cfginfo': BuildConfigInfo,
    # 'build_py': DataladBuild
}

# PyPI doesn't render markdown yet. Workaround for a sane appearance
# https://github.com/pypa/pypi-legacy/issues/148#issuecomment-227757822
README = opj(dirname(__file__), 'README.md')
try:
    import pypandoc
    long_description = pypandoc.convert(README, 'rst')
except ImportError:
    long_description = open(README).read()

setup(
    name="niceman",
    author="The NICEMAN Team and Contributors",
    author_email="team@niceman.org",
    version=version,
    description="Neuroimaging Computational Environments Manager",
    long_description=long_description,
    packages=niceman_pkgs,
    install_requires=requires['core'],
    extras_require=requires,
    entry_points={
        'console_scripts': [
            'niceman=niceman.cmdline.main:main',
        ],
    },
    cmdclass=cmdclass,
    package_data={
        'niceman': []
    }
)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Support for Redhat(-based) distribution(s)."""
import attr

from six.moves import map

import datetime
import logging
import re

from niceman.distributions.base import DistributionTracer

lgr = logging.getLogger('niceman.distributions.redhat')

from .base import SpecObject
from .base import Package
from .base import Distribution
from .base import TypedList
from .base import _register_with_representer
from ..support.exceptions import CommandError
from ..utils import attrib


@attr.s(cmp=True)
class RPMSource(SpecObject):
    """RPM origin information"""
    id = attrib(default=attr.NOTHING)
    name = attrib()
    revision = attrib()
    updated = attrib()
    pkgs = attrib()
    size = attrib()
    mirrors = attrib()
    metalink = attrib()
    baseurl = attrib()
    expire = attrib()
    filename = attrib()


_register_with_representer(RPMSource)


@attr.s(slots=True, frozen=True, cmp=False, hash=True)
class RPMPackage(Package):
    """Redhat package information"""
    name = attrib(default=attr.NOTHING)
    pkgid = attrib()
    version = attrib()
    release = attrib()
    architecture = attrib()
    install_date = attrib()
    group = attrib()
    size = attrib()
    license = attrib()
    signature = attrib()
    source_rpm = attrib()
    build_date = attrib()
    build_host = attrib()
    packager = attrib()
    vendor = attrib()
    url = attrib()
    files = attrib(default=attr.Factory(list), hash=False)

    def satisfies(self, other):
        """return True if this package (self) satisfies the requirements of 
        the passed package (other)"""
        if not isinstance(other, Package):
            raise TypeError('satisfies() requires a package argument')
        if not isinstance(other, RPMPackage):
            return False
        if self.name != other.name:
            return False
        if other.version is not None and self.version != other.version:
            return False
        if other.architecture is not None \
                and self.architecture != other.architecture:
            return False
        return True


_register_with_representer(RPMPackage)


@attr.s
class RedhatDistribution(Distribution):
    """
    Class to provide Redhat-based shell commands.
    """

    sources = TypedList(RPMSource)
    packages = TypedList(RPMPackage)
    version = attrib()  # version as depicted by /etc/redhat_version

    def initiate(self, session):
        """
        Perform any initialization commands needed in the environment environment.

        Parameters
        ----------
        environment : object
            The Environment sub-class object.
        """
        lgr.debug("Adding Redhat update to environment command list.")
        self._init_repo_sources(session)

    def _init_repo_sources(self, session):
        """
        Install package repos listed in the spec but missing in the resource.

        Parameters
        ----------
        session : Session object
        """

        # TODO: Expand the listing of 3rd party repos as popular ones in the
        # neuroscience community become apparent.
        third_party_repos = {
            'epel/x86_64': 'epel-release'
        }

        # Get a list of the enabled repos on the system.
        out, _ = session.execute_command(['yum', 'repolist'])
        system_repos = []
        skip = True
        for line in out.splitlines():
            if line.startswith('repo id'):
                skip = False
                continue
            if skip:
                continue
            system_repos.append(line.split()[0])

        # Try to install any repos listed in the spec that are missing on the system.
        for source in self.sources:
            if source.id not in system_repos:
                if source.id in third_party_repos:
                    session.execute_command(['yum', 'install', '-y',
                        third_party_repos[source.id]])
                else:
                    lgr.error("Unable to install source repo \"%s\" found in "
                        "spec but not in resource environment.", source.id)

    def install_packages(self, session, use_version=True):
        """
        Install the packages associated to this distribution by the provenance
        into the environment.

        Parameters
        ----------
        session : object
        use_version : bool, optional
          Use version information if provided.
        """
        package_specs = []

        for package in self.packages:
            if use_version:
                package_spec = package.pkgid
            else:
                package_spec = package.name
            package_specs.append(package_spec)

        # Doing in one shot to fail early if any of the versioned specs
        # couldn't be satisfied
        lgr.debug("Installing %s", ', '.join(package_specs))
        session.execute_command(
            # TODO: Pull env out of provenance for this command.
            ['yum', 'install', '-y'] + package_specs,
            # env={'DEBIAN_FRONTEND': 'noninteractive'}
        )

    def satisfies_package(self, package):
        """return True if this distribution (self) satisfies the requirements 
        of the passed package"""
        if not isinstance(package, Package):
            raise TypeError('satisfies_package() requires a package argument')
        if not isinstance(package, RPMPackage):
            return False
        return any([p.satisfies(package) for p in self.packages])

    def satisfies(self, other):
        """return True if this distribution (self) satisfies the requirements 
        of the other distribution (other)"""
        if not isinstance(other, Distribution):
            raise TypeError('satisfies() requires a distribution argument')
        if not isinstance(other, RedhatDistribution):
            return False
        return all(map(self.satisfies_package, other.packages))

    def __sub__(self, other):
        # the semantics of distribution subtraction are, for d1 - d2:
        #     what is specified in d1 that is not specified in d2
        #     or how does d2 fall short of d1
        #     or what is in d1 that isn't satisfied by d2
        return [p for p in self.packages if not other.satisfies_package(p)]


_register_with_representer(RedhatDistribution)


class RPMTracer(DistributionTracer):
    """.rpm-based (and using yum and rpm) systems package tracer
    """

    # The Redhat tracer is not designed to handle directories
    HANDLES_DIRS = False

    def _init(self):
        # TODO: we might want a generic helper for collections of things
        # where we could match based on the set of attrs which matter
        self._package_install_dates = {}

    def identify_distributions(self, files):
        """
        Return a distribution object containing package and repos source
        information for files found to be from CentOS distribution

        Parameters
        ----------
        files : list(string)
            List of files to check to see if they come from CentOS packages

        Returns
        -------
        dist : RedhatDistribution object
        remaining_files : list(string)
            List of files not identified to be in CentOS packages
        """
        if not files:
            return

        try:
            redhat_version = self._session.read('/etc/redhat-release').strip()
            _, _ = self._session.execute_command('ls -ld /etc/yum')
        except CommandError as exc:
            lgr.debug("Did not detect Redhat (or derivative): %s", exc)
            return

        packages, remaining_files = self.identify_packages_from_files(files)
        # TODO: add option to report distribution even if no packages/files
        # found association
        if not packages:
            return

        dist = RedhatDistribution(
            name="redhat",
            version=redhat_version,
            packages=packages,
            sources=self._find_all_sources()
        )

        yield dist, remaining_files

    def _get_packagefields_for_files(self, files):
        """
        Query the system for detail information for each package found.

        Parameters
        ----------
        files : list(string)
            List of files to trace which packages they come from

        Returns
        -------
        dictionary : key = package id, value = dict of package details
        """

        file_to_package_dict = {}

        # Get a list of the attrs in the RPMPackage class to filter
        # out non-relavant values retrieved from the system.
        package_fields = [p.name for p in attr.fields(RPMPackage)]

        for file in files:
            try:
                # Get the package identifier that the file is a member.
                pkgid, err = self._session.execute_command(['rpm', '-qf',
                    file])
            except CommandError:
                continue
            if err:
                continue

            pkgids = pkgid.splitlines()
            if len(pkgids) > 1:
                msg = "Multiple packages found for file {}: {}. Selecting {}"
                lgr.info(msg.format(file, ', '.join(pkgids), pkgids[0]))
            pkgid = pkgids[0].strip()

            # Get the package information from the system.
            package_info, _ = self._session.execute_command(['rpm', '-qi',
                pkgid])
            if package_info:
                # Store the relevant fields of the package in the dict.
                # The labels of the fields returned by the system are mapped
                # to the attr fields in the RPMPacakge class.
                pkg = {'pkgid': pkgid}
                for line in package_info.splitlines():
                    matches = re.match(r'(.*?):(.*)\s{2,}(.*):(.*)', line)
                    if not matches:
                        matches = re.match(r'(.*?):(.*)', line)
                    if matches:
                        for i in range(1, len(matches.groups()), 2):
                            key = matches.group(i).strip().lower().replace(' ',
                                '_')
                            if key in package_fields:
                                pkg[key] = matches.group(i+1).strip()
                lgr.debug("Identified file %r to belong to package %s",
                          pkgid, pkg)
                file_to_package_dict[file] = pkg
        return file_to_package_dict

    def _create_package(self, name, **kwargs):
        return RPMPackage(name=name, **kwargs)

    def _find_all_sources(self):
        """
        Retrieve repository source information from the system

        Returns
        -------
        sources : list(RPMSource)
        """
        attr_fields = {f.name for f in attr.fields(RPMSource)}
        sources = []
        # Get all repo info from the system and store information for each
        # enabled repo in a RPMSource object.
        out, _ = self._session.execute_command(['yum', '--verbose',
            'repolist'])
        for line in out.splitlines():
            if line.startswith('Repo-'):
                key, value = line.split(':', 1)
                key = key.strip()
                value = value.strip()
                if key in ['Repo-pkgs', 'Repo-size']:
                    continue
                if key == ('Repo-id'):
                    values = {'id': value}
                elif key in ['Repo-pkgs', 'Repo-size']:
                    continue
                elif key == 'Repo-expire':
                    delta, last = re.match(r'(\d+) second\(s\) \(last: (.*)\)',
                        value).groups()
                    values['expire'] = (datetime.datetime.strptime(last, "%c")
                        + datetime.timedelta(0, int(delta))).strftime("%c")
                elif key == 'Repo-baseurl':
                    values['baseurl'] = re.match(r'(\S+)', value).groups()[0]
                else:
                    # Map the field labels returned by the sytem to the attr
                    # fields in the RPMSource class.
                    field = key.split('-')[1]
                    if field in attr_fields:
                        values[field] = value
                    else:
                        lgr.warning("Ignoring RPM line: %s", line)
            if len(line) == 0:
                sources.append(RPMSource(**values))
        return sources



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Orchestrator sub-class to provide management of the localhost environment."""

from niceman.distributions.debian import DebianDistribution

import logging
lgr = logging.getLogger('niceman.distributions.neurodebian')

class NeurodebianDistribution(DebianDistribution):
    """
    Class to provide Debian-based shell commands.
    """

    def __init__(self, provenance):
        """
        Class constructor

        Parameters
        ----------
        provenance : dictionary
            Provenance information for the distribution.
        """
        super(NeurodebianDistribution, self).__init__(provenance)


    def initiate(self, environment):
        """
        Perform any initialization commands needed in the environment.

        Parameters
        ----------
        environment : object
            The Environment sub-class object.
        """

        # TODO: Add code to setup NeuroDebian repository setup.

        super(NeurodebianDistribution, self).initiate(environment)